import aiohttp
import pytest

import rnc.corpora_requests as req
from rnc.corpora import RNC_URL


correct_params = {
    'env': 'alpha',
    'api': '1.0',
    'lang': 'ru',
    'mode': 'main',
    'text': 'lexgramm',
    'out': 'normal',
}
wrong_params = {
    'shue': 'ppsh'
}


def test_wrong_range():
    with pytest.raises(ValueError):
        req.get_htmls(RNC_URL, 10, 0)


def test_wrong_params():
    with pytest.raises(aiohttp.ClientResponseError):
        req.get_htmls(RNC_URL, **wrong_params)


def test_incorrect_request():
    with pytest.raises(ValueError):
        req.is_request_correct(RNC_URL, 1, **wrong_params)


def test_correct_request():
    correct_params['lex1'] = 'корпус'
    res = req.is_request_correct(RNC_URL, 1, **correct_params)
    assert res is True


def test_no_results_found():
    wrong_params = correct_params
    wrong_params['lex1'] = 'ssssssssss'

    res = req.whether_result_found(RNC_URL, **wrong_params)
    assert res is False


def test_results_found():
    correct_params['lex1'] = 'ты'
    res = req.whether_result_found(RNC_URL, **correct_params)
    assert res is True


def test_page_does_not_exist():
    wrong_params = correct_params
    wrong_params['lex1'] = 'васкуляризация'

    res = req.does_page_exist(RNC_URL, 10, **wrong_params)
    assert res is False


def test_page_exists():
    correct_params['lex1'] = 'я'
    res = req.does_page_exist(RNC_URL, 20, **correct_params)
    assert res is True


def test_wait_some_time():
    correct_params['lex1'] = 'я'
    html_codes = req.get_htmls(RNC_URL, 0, 15, **correct_params)
    assert len(html_codes) is 15



__all__ = (
    'MainCorpus',
    'DialectalCorpus',
    'AccentologicalCorpus',
    'SpokenCorpus',
    'PaperRegionalCorpus',
    'Paper2000Corpus',
    'ParallelCorpus',
    'MultilingualParaCorpus',
    'TutoringCorpus',
    'MultimodalCorpus'
)

import csv
import json
import logging
import os
import random
import re
import string
import time
import webbrowser
from pathlib import Path
from typing import Dict, Callable, List, Any, Tuple
from urllib.parse import unquote

import bs4

import rnc.corpora_requests as creq
import rnc.examples as expl

logger = logging.getLogger("rnc")


# Russian National Corpus URL
RNC_URL = "https://processing.ruscorpora.ru/search.xml"
BASE_RNC_URL = "https://processing.ruscorpora.ru"

ALPHABET = f"{string.ascii_letters}{string.digits}"


def create_filename(length: int = 8) -> str:
    """ Create random filename.

    :param length: int, length of result (8 by default).
    :return: str, random symbols.
    """
    name = random.sample(ALPHABET, length)
    return ''.join(name)


def create_unique_filename(folder: Path,
                           class_name: str,
                           p_count: int) -> Path:
    """ Create a random unique csv filename,
    means the file does not exist.

    Name format:
    <class_name><p_count>_<some random symbols>

    :param folder: Path to the data folder.
    :param class_name: name of the class.
    :param p_count: int, count of pages.
    :return: Path, unique filename.
    """
    name_template = "{}{}_{}.csv"
    name = name_template.format(class_name, p_count, create_filename())
    path = folder / name
    while path.exists():
        name = name_template.format(class_name, p_count, create_filename())
        path = path.with_name(name)
    return path


def clean_text_up(text: str) -> str:
    """ Remove duplicate spaces from str and strip it.

    :param text: str to clean up.
    :return: str, clean str.
    """
    return ' '.join(text.split()).strip()


def create_doc_url(doc_url: str) -> str:
    """ Create full url to document in RNC.
    Add https://... to doc.

    :param doc_url: str, doc url to extend.
    :return: str, extended url.
    """
    if not doc_url:
        return doc_url
    return f"{BASE_RNC_URL}/{doc_url}"


def join_with_plus(item: str) -> str:
    """ Split txt and join it with '+'.

    :param item: str, text to convert.
    :return: str, converted text.
    """
    res = item.split()
    return '+'.join(res)


def str_to_int(value: str) -> int:
    """ Convert str like '350 000 134' to int.

    :param value: str to convert.
    :return: int.
    """
    return int(value.replace(' ', ''))


class Corpus:
    """ Base class for Corpora """
    # default params
    # documents per page
    __DPP = 5
    # sentences per document
    __SPD = 10
    # search format
    __TEXT = 'lexgramm'
    # output format
    __OUT = 'normal'
    # accent on words, without accent by default
    __ACCENT = '0'
    # show order
    __SORT = 'i_grtagging'
    # distance between n and (n + 1) words
    __MIN = 1
    __MAX = 3
    # count of examples to print
    __RESTRICT_SHOW = 50

    # symbols to write csv
    _DATA_W_DELIMITER = '\t'
    _DATA_W_QUOTCHAR = '"'

    DATA_FOLDER = Path('data')

    def __init__(self,
                 query: dict or str = None,
                 p_count: int = None,
                 file: str or Path = None,
                 **kwargs) -> None:
        """ 
        If the file exists, working with a local database.

        :param query: dict of str, words to search;
         {word1: {properties}, word2: {properties}...}.
         If you chose 'lexform' as a 'text' param, you must give here a string.
        :param p_count: int, count of pages to request.
        :param file: str or Path, filename of a local database.
        :keyword dpp: str or int, documents per page.
        :keyword spd: str or int, sentences per document.
        :keyword text: str, search format: 'lexgramm' or 'lexform'.
        :keyword out: str, output format: 'normal' or 'kwic'.
        :keyword kwsz: str or int, count of words in context;
         Optional param if 'out' is 'kwic'.
        :keyword sort: str, sort show order. See docs how to set it.
        :keyword subcorpus: str, subcorpus. See docs how to set it.
        :keyword expand: str, if 'full', all part of doc will be shown.
        :keyword accent: str or int, with accents on words or not:
         1 – with, 0 – without.

        :keyword marker: function, with which found words will be marked.
        """
        # list of examples
        self._data = []
        # http tags to request
        self._params = {}
        # found wordforms with their frequency
        self._found_wordforms = {}
        # query, wordforms to found
        self._query = {}
        # count of PAGES
        self._p_count = 0
        # type of example should be defined before params init
        self._ex_type = kwargs.pop('ex_type', None)
        self._marker = kwargs.pop('marker', None)
        # additional info from the first page:
        # amount of docs, contexts, where the query was found,
        # link to the graphic with distribution by years
        self._add_info = {}

        # path to local database
        class_name = self.__class__.__name__.replace('Corpus', '')
        file = file or create_unique_filename(
            self.DATA_FOLDER, class_name, p_count)
        path = Path(file)
        if path.suffix != '.csv':
            msg = "File must have '.csv' extension"
            logger.error(msg)
            raise TypeError(msg)

        # to these files the data and req params will be dumped
        self._csv_path = path
        self._config_path = path.with_suffix('.json')

        # init from file if it exists
        if self._csv_path.exists():
            try:
                self._from_file()
            except FileExistsError:
                logger.exception('')
                raise
        # or working with RNC
        else:
            self._from_corpus(query, p_count, **kwargs)

    def _from_corpus(self,
                     query: dict or str,
                     p_count: int,
                     **kwargs) -> None:
        """ Init from the given values. If the file does not exist.
        Params the same as in the init method.

        :return: None.
        """
        if not query:
            msg = "Query must be not empty"
            logger.error(msg)
            raise ValueError(msg)
        self._query = query

        if p_count <= 0:
            msg = "Page count must be > 0"
            logger.error(msg)
            raise ValueError(msg)
        self._p_count = p_count

        # base params
        self._params['env'] = 'alpha'
        self._params['api'] = '1.0'
        self._params['lang'] = 'en'

        self._params['dpp'] = kwargs.pop('dpp', None) or self.__DPP
        self._params['spd'] = kwargs.pop('spd', None) or self.__SPD
        self._params['text'] = kwargs.pop('text', None) or self.__TEXT
        self._params['out'] = kwargs.pop('out', None) or self.__OUT
        self._params['sort'] = kwargs.pop('sort', None) or self.__SORT

        accent = kwargs.pop('accent', None) or self.__ACCENT
        accent = int(accent)
        self._params['nodia'] = int(not accent)

        if self.params['out'] == 'kwic' and 'kwsz' in kwargs:
            self._params['kwsz'] = kwargs.pop('kwsz')

        if 'subcorpus' in kwargs:
            mycorp = kwargs.pop('subcorpus')
            self._params['mycorp'] = unquote(mycorp)

        # TODO: page structure changed if expand=full
        # if 'expand' in kwargs:
        #     self._params['expand'] = kwargs.pop('expand')

        self._query_to_http()

        # parsing depends on 'out' value
        self._page_parser = None
        self._page_parser_and_ex_type()

    def _from_file(self) -> None:
        """ Load data and params from the local databases.
        If the file exists.

        :return: None.
        """
        if not (self._csv_path.exists() and self._config_path.exists()):
            raise FileExistsError("Data and config file must exist together")

        params = self._load_params()
        self._query = params.get('query', None)
        self._p_count = params.get('p_count', None)
        self._params = params.get('params', None)

        mode = self.mode
        if mode is None or mode != self._MODE:
            msg = f"Tried to load data to wrong Corpus: " \
                  f"{mode} instead of {self._MODE}"
            logger.error(msg)
            raise NotImplementedError(msg)

        # these params must be defined here too
        self._page_parser_and_ex_type()

        self._data = self._load_data()
        # add info about
        try:
            self._get_additional_info()
        except Exception:
            logger.exception("It is impossible to get additional info from RNC")

    def _load_data(self) -> List:
        """ Load data from csv file.

        :return: list of examples.
        """
        with self.file.open('r', encoding='utf-16') as f:
            dm = self._DATA_W_DELIMITER
            qch = self._DATA_W_QUOTCHAR
            reader = csv.reader(f, delimiter=dm, quotechar=qch)
            # first row contains headers, skip it
            next(reader)

            data = [self.ex_type(*row) for row in reader]
            wordforms = list(map(lambda example: example.found_wordforms, data))
            wordforms = sum(wordforms, [])
            self._add_wordforms(wordforms)

        return data

    def _load_params(self) -> Dict:
        """ Load request params from json file.

        :return: json dict.
        """
        with self._config_path.open('r', encoding='utf-16') as f:
            return json.load(f)

    @classmethod
    def set_dpp(cls, value: int) -> None:
        if not isinstance(value, int):
            logger.error("DPP must be int")
            raise TypeError("DPP must be int")
        cls.__DPP = value

    @classmethod
    def set_spd(cls, value: int) -> None:
        if not isinstance(value, int):
            logger.error("SPD must be int")
            raise TypeError("SPD must be int")
        cls.__SPD = value

    @classmethod
    def set_text(cls, value: str) -> None:
        if not isinstance(value, str):
            logger.error("Text must be str")
            raise TypeError("Text must be str")
        cls.__TEXT = value

    @classmethod
    def set_sort(cls, value: str) -> None:
        if not isinstance(value, str):
            logger.error("Sort key must be str")
            raise TypeError("Sort key must be str")
        cls.__SORT = value

    @classmethod
    def set_min(cls, value: int) -> None:
        if not isinstance(value, int):
            logger.error("min must be int")
            raise TypeError("min must be int")
        cls.__MIN = value

    @classmethod
    def set_max(cls, value: int) -> None:
        if not isinstance(value, int):
            logger.error("max must be int")
            raise TypeError("max must be int")
        cls.__MAX = value

    @classmethod
    def set_restrict_show(cls, value: int or bool) -> None:
        if not isinstance(value, (int, bool)):
            logger.error("Restrict count must be int or bool")
            raise TypeError("Restrict count must be int or bool")
        cls.__RESTRICT_SHOW = value

    @staticmethod
    def _get_ambiguation(tag: bs4.element.Tag) -> str:
        """ Get pretty ambiguation from example.

        :param tag: bs4.element.Tag, example.
        :return: str, 'disambiguated' or 'not disambiguated' or 'Not found'.
        """
        ambiguation = (tag.find('span', {'class': 'on'}) or
                       tag.find('span', {'class': 'off'}))
        if not ambiguation:
            return 'Not found'
        ambiguation = ambiguation.text.strip()
        # here ambiguation like '[...]'
        ambiguation = ambiguation[1:-1].strip()
        return ambiguation

    @staticmethod
    def _get_text(tag: bs4.element.Tag) -> str:
        """ Get pretty text from example and remove
        from there duplicate spaces.

        Here it is assumed, that all examples have text.

        :param tag: bs4.element.Tag, example.
        :return: str, text.
        """
        # using 'findall' method removes punctuation marks
        txt = tag.get_text()
        # remove duplicate spaces
        return clean_text_up(txt)

    @staticmethod
    def _get_doc_url(tag: bs4.element.Tag) -> str:
        """ Get pretty doc url from example.

        :param tag: bs4.element.Tag, example.
        :return: str, doc url or 'Not found'.
        """
        doc_url = tag.a
        if not doc_url:
            return 'Not found'
        doc_url = doc_url.attrs['href']
        return create_doc_url(doc_url)

    @staticmethod
    def _get_source(tag: bs4.element.Tag) -> str:
        """ Get pretty source from example.

        :param tag: bs4.element.Tag, example.
        :return: str, examples source or 'Not found'.
        """
        src = tag.find('span', {'class': 'doc'})
        if not src:
            return "Not found"
        src = clean_text_up(src.text)
        # here src like '[...]'
        src = src[1:-1].strip()
        return src

    @staticmethod
    def _parse_lexgramm_params(params: dict or str,
                               join_inside_symbol: str,
                               with_braces: bool = False) -> str:
        """ Convert lexgramm params to str for HTTP request.

        :param params: dict, params to convert.
        :param join_inside_symbol: str, symbol to join params.
        :param with_braces: bool, whether the braces 
        will be added around the param.
        :return: str, joined with ',' params.
        """
        if not (isinstance(params, (str, dict)) and ' ' not in params):
            msg = f"Param must be str without spaces or dict, given: {params}"
            logger.error(msg)
            raise ValueError(msg)
        # let the user to give only one param:
        # {'word': {'gramm': 'acc', 'flags': 'bmark'}}
        if isinstance(params, str):
            return params

        res = []
        for val in params.values():
            if not isinstance(val, (str, int, list)):
                msg = "One should give to tags only str, list or int"
                logger.error(msg)
                raise ValueError(msg)

            if isinstance(val, (str, int)):
                val = [str(val)]

            item = f"{'(' * with_braces}" \
                   f"{join_inside_symbol.join(val)}" \
                   f"{')' * with_braces}"
            res += [item]
        return ','.join(res)

    @staticmethod
    def _find_searched_words(tag: bs4.element.Tag) -> List[str]:
        """ Get searched words from tag, they are marked with 'g-em'
        parameter in the class name. Strip them.

        :param tag: bs4.element.Tag, tag with result.
        :return: list of string, words to which request was.
        """
        # TODO: simplify
        # params of the classes and word if 'class' is
        class_params = [
            (i.attrs.get('class', ''), i.text)
            for i in tag.contents
            if isinstance(i, bs4.element.Tag)
        ]
        # searched words are marked by class parameter 'g-em'
        searched_words = [
            i[1].strip()
            for i in class_params
            if 'g-em' in i[0]
        ]
        return searched_words

    @property
    def data(self) -> List:
        """
        :return: list of examples.
        """
        return self._data

    @property
    def query(self) -> Dict[str, dict] or str:
        """
        :return: dict or str, requested items.
        """
        return self._query

    @property
    def forms_in_query(self) -> List[str]:
        """
        :return: list of str, requested words.
        """
        req = self.query
        if isinstance(req, str):
            return req.split()
        return list(req.keys())

    @property
    def p_count(self) -> int:
        """
        :return: int, requested count of pages.
        """
        return self._p_count

    @property
    def file(self) -> Path:
        """
        :return: Path, path to local database file.
        """
        return self._csv_path

    @property
    def marker(self) -> Callable:
        """
        :return: function to mark found wordforms.
        """
        return self._marker

    @property
    def params(self) -> dict:
        """
        :return: dict of HTTP params.
        """
        return self._params

    @property
    def found_wordforms(self) -> dict:
        """ Get info about found wordforms, {form: frequency}.

        :return: dict of str, found wordforms and their frequency.
        """
        return self._found_wordforms

    @property
    def url(self) -> str:
        """ Return URL, first page of RNC results.

        :return: str, URL.
        """
        params = '&'.join(
            f"{key}={val}"
            for key, val in self.params.items()
        )
        return f"{RNC_URL}?{params}"

    @property
    def ex_type(self) -> Any:
        """
        :return: example type of the Corpus.
        """
        return self._ex_type

    @property
    def amount_of_docs(self) -> int or None:
        """ Get amount of documents, where the query was found.

        :return: int, this amount or None if it does not exist.
        """
        return self._add_info.get('docs', None)

    @property
    def amount_of_contexts(self) -> int or None:
        """ Get amount of contexts, where the query was found.

        :return: int, this amount or None if it does not exist.
        """
        return self._add_info.get('contexts', None)

    @property
    def graphic_link(self) -> str or None:
        """ Get the link the graphic of the
        distribution of query occurrences by years.

        :return: str, this link or None if it does not exist.
        """
        return self._add_info.get('graphic_link', None)

    def open_graphic(self) -> None:
        """ Open the graph of the distribution
        of query occurrences by years.

        :return: None.
        """
        url = self.graphic_link
        if url is None:
            msg = "Graphic does not exist"
            logger.error(msg)
            raise RuntimeError(msg)

        try:
            webbrowser.open_new_tab(url)
        except Exception:
            logger.exception("It is impossible to open the graphic")
            raise

    @staticmethod
    def _get_where_query_found(content: bs4.element.Tag) -> Dict[str, Any]:
        """ Get converted to int amount of found docs and contexts.

        :param content: bs4.element.Tag, here these values are.
        """
        res = {}
        amount = content.find('p', {'class': 'found'})
        blocks = amount.find_all('span', {'class': 'stat-number'})

        contexts = blocks[-1].get_text()
        res['contexts'] = str_to_int(contexts)
        if len(blocks) is 2:
            docs = blocks[0].get_text()
            res['docs'] = str_to_int(docs)
        return res

    @staticmethod
    def _get_graphic_url(content: bs4.element.Tag) -> str or None:
        """ Get distribution by years graphic URL.

        :param content: bs4.element.Tag, here the link is.
        :return: str, full URL.
        """
        a = content.find('a', {'target': '_blank'})
        try:
            link = a['href']
        except Exception:
            return
        return f"{BASE_RNC_URL}/{link}"

    def _get_additional_info(self) -> None:
        """ Get additional info (amount of found docs and contexts,
        link to graphic with distribution by years).

        :return: None.
        """
        params = self.params.copy()
        params['lang'] = 'ru'
        params.pop('expand', None)
        try:
            first_page_code = creq.get_htmls(RNC_URL, **params)[0]
        except Exception:
            raise

        soup = bs4.BeautifulSoup(first_page_code, 'lxml')
        content = soup.find('div', {'class': 'content'})

        additional_info = Corpus._get_where_query_found(content)
        graphic_url = Corpus._get_graphic_url(content)
        if graphic_url:
            additional_info['graphic_link'] = graphic_url

        self._add_info = additional_info.copy()

    def _page_parser_and_ex_type(self) -> None:
        """ Add 'parser' and 'ex_type' params.
        They are depended on 'out' tag.

        :return: None
        """
        if self.out == 'normal':
            # ex_type is defined above in this case
            self._page_parser = self._parse_page_normal
        elif self.out == 'kwic':
            self._page_parser = self._parse_page_kwic
            self._ex_type = expl.KwicExample

    def _query_to_http(self) -> None:
        """ Convert the query to HTTP tags, add them to params.

        :return: None.
        :exception ValueError: if wrong type given.
        """
        if self.text == 'lexform':
            if not isinstance(self.query, str):
                msg = "One must give str as a query if search is 'lexform'"
                logger.error(msg)
                raise ValueError(msg)
            self._params['req'] = join_with_plus(self.query)

        # in lexgramm search one word may be too
        if isinstance(self.query, str):
            # working with long query like
            # 'открыть -открыл дверь -двери настеж'
            q = re.finditer(r'\b\w+\b( -\b\w+\b)?', self.query)
            for num, query in enumerate(q, 1):
                match = query.group(0)
                self._params[f"lex{num}"] = join_with_plus(match)
                if num > 1:
                    self._params[f"min{num}"] = self.__MIN
                    self._params[f"max{num}"] = self.__MAX
            return

        # words and their params
        for word_num, (word, params) in enumerate(self.query.items(), 1):
            # add distance
            if word_num > 1:
                min_distance = f'min{word_num}'
                max_distance = f'max{word_num}'
                # given or default values
                if isinstance(params, dict):
                    self._params[min_distance] = params.get('min', None) or self.__MIN
                    self._params[max_distance] = params.get('max', None) or self.__MAX
                else:
                    self._params[min_distance] = self.__MIN
                    self._params[max_distance] = self.__MAX

            self._params[f"lex{word_num}"] = join_with_plus(word)

            if isinstance(params, str):
                # empty param, skip it
                continue

            # grammar properties
            gramm = params.get('gramm', '')
            if gramm:
                try:
                    gram_props = Corpus._parse_lexgramm_params(gramm, '|', True)
                except Exception:
                    raise
                self._params[f"gramm{word_num}"] = gram_props

            # additional properties
            flags = params.get('flags', '')
            if flags:
                try:
                    flag_prop = Corpus._parse_lexgramm_params(flags, '+')
                except Exception:
                    raise
                self._params[f"flags{word_num}"] = flag_prop

            # TODO: semantic properties
            sem = params.get('sem', '')
            if sem:
                try:
                    logger.warning("Semantic properties does not support")
                except Exception:
                    raise
                # self.__params фильтр1 и фильтр2

    def _add_wordforms(self,
                       forms: List[str]) -> None:
        """ Add found wordforms to counter. Low and strip items.

        :param forms: list of str, wordforms to add.
        :return: None.
        """
        if not forms:
            return

        for form in forms:
            form = clean_text_up(form).lower()
            self._found_wordforms[form] = self.found_wordforms.get(form, 0) + 1

    def _parse_doc(self,
                   doc: bs4.element.Tag) -> Any:
        """ Parse the doc to list of Examples.

        Parsing depends on the subcorpus,
        the method redefined at the descendants.
        """
        # TODO: remake this func to generator?
        msg = "The func not implemented to the parent Corpus class"
        logger.error(msg)
        raise NotImplementedError(msg)

    def _parse_example(self,
                       *args,
                       **kwargs) -> Any:
        """ Parse the example to Example object.

        Parsing depends on the subcorpus,
        the method redefined at the descendants.
        """
        msg = "The func not implemented to the parent Corpus class"
        logger.error(msg)
        raise NotImplementedError(msg)

    def _parse_kwic_example(self,
                            left: bs4.element.Tag,
                            center: bs4.element.Tag,
                            right: bs4.element.Tag) -> expl.KwicExample:
        l_txt = clean_text_up(left.text)
        c_txt = clean_text_up(center.text)
        # remove ←…→ symbol too
        r_txt = clean_text_up(right.text)[:-4].rstrip()

        found_wordforms = Corpus._find_searched_words(left)
        found_wordforms += Corpus._find_searched_words(center)
        found_wordforms += Corpus._find_searched_words(right)

        try:
            src = right.a.attrs['msg'].strip()
            url = right.a.attrs['href']
        except Exception:
            logger.exception("Source or url not found")
            src = url = ''

        url = create_doc_url(url)

        new_ex = expl.KwicExample(
            l_txt, c_txt, r_txt, src, found_wordforms, url)
        new_ex.mark_found_words(self.marker)

        return new_ex

    def _parse_page_kwic(self,
                         page: str) -> List[expl.KwicExample]:
        """ Parse page if 'out' is 'kwic'.

        :param page: str, html code of page to parse.
        :return: list of examples.
        :exception ValueError: if the content not found.
        """
        soup = bs4.BeautifulSoup(page, 'lxml')
        res = []

        content = soup.find('table', {'align': 'left'})
        if not content:
            msg = "Content is None, this behavior is undefined, contact the developer"
            logger.error(msg)
            raise ValueError(msg)

        nobr = content.find_all('nobr')
        if len(nobr) % 3:
            logger.warning("Len of nobr tags list % 3 != 0")

        for left, center, right in zip(nobr[::3], nobr[1::3], nobr[2::3]):
            new_ex = self._parse_kwic_example(left, center, right)
            res += [new_ex]
            self._add_wordforms(new_ex.found_wordforms)
        return res

    def _parse_page_normal(self,
                           page: str) -> List:
        """ Parse page if 'out' is 'normal'.

        :param page: str, html code to parse.
        :return: list of examples.
        """
        soup = bs4.BeautifulSoup(page, 'lxml')
        res = []

        for doc in soup.find_all('li'):
            try:
                parsed_doc = self._parse_doc(doc)
            except Exception:
                logger.exception("Error while parsing doc")
            else:
                res += parsed_doc

        return res

    def _parse_all_pages(self,
                         pages: List[str]) -> List:
        """ Parse all pages.

        :param pages: list of str, html codes of the pages.
        :return: list of examples.
        """
        parsed = [self._page_parser(page) for page in pages]
        return sum(parsed, [])

    def _data_to_csv(self) -> None:
        """ Dump the data to csv file.
        Here it is assumed that the data exist.

        :return: None.
        """
        data = [example.items for example in self.data]
        columns = self[0].columns
        with self.file.open('w', encoding='utf-16', newline='') as f:
            # class constants
            dm = self._DATA_W_DELIMITER
            qch = self._DATA_W_QUOTCHAR

            writer = csv.writer(
                f, delimiter=dm, quotechar=qch, quoting=csv.QUOTE_MINIMAL)
            writer.writerows([columns] + data)

    def _params_to_json(self) -> None:
        """ Write the request params: query,
        p_count and http tags to json file.

        Here it is assumed that these params exist.

        :return: None.
        """
        to_write = {
            'query': self.query,
            'p_count': self.p_count,
            'params': self.params
        }
        with self._config_path.open('w', encoding='utf-16') as f:
            json.dump(to_write, f, indent=4, ensure_ascii=False)

    def dump(self) -> None:
        """ Write the data to csv file, request params to json file.

        :return: None.
        :exception RuntimeError: If there are no data, params or files exist.
        """
        if not self.data:
            logger.error("There is no data to write")
            raise RuntimeError("There is no data to write")
        if not (self.query and self.p_count and self.params):
            logger.error("There is no data to write")
            raise RuntimeError("There is no data to write")

        os.makedirs(self.DATA_FOLDER, exist_ok=True)

        self._data_to_csv()
        self._params_to_json()

        logger.debug(
            f"Data was wrote to files: {self.file} and {self._config_path}")

    def open_url(self) -> None:
        """ Open first page of RNC results in the new
        tab of the default browser.

        :return: None.
        :exception ValueError: if url is wrong.
        :exception: if sth went wrong.
        """
        try:
            webbrowser.open_new_tab(self.url)
        except Exception:
            logger.exception(
                f"Error while opening doc with url: {self.url}")
            raise

    def request_examples(self) -> None:
        """ Request examples, parse them and update the data.

        If there are no results found, last page does not exist,
        params or query is wrong then exception.

        :return: None.

        :exception RuntimeError: if the data still exist.
        """
        if self.data:
            logger.error("Tried to request new examples, however data exist")
            raise RuntimeError("Data still exist")

        try:
            # TODO: get first page code if everything is OK
            creq.is_request_correct(RNC_URL, self.p_count, **self.params)
        except Exception:
            msg = f"Query = {self.forms_in_query}, {self.p_count}, {self.params}"
            logger.exception(msg)
            raise

        # get additional info from the first RNC page.
        logger.info("Requesting additional info from the first RNC page")
        self._get_additional_info()
        logger.info("Additional info was successfully received")

        logger.info("Main request")
        coro_start = time.time()
        htmls = creq.get_htmls(RNC_URL, 0, self.p_count, **self.params)
        logger.info("Main request was successfully completed")
        logger.info(f"Coro executing time: {time.time() - coro_start:.2f}")

        logger.info("Parsing html was started")
        try:
            parsing_start = time.time()
            parsed = self._parse_all_pages(htmls)
            parsing_stop = time.time()
        except Exception:
            logger.exception(f"Error while parsing, query = {self.params}")
            raise
        else:
            logger.info("Parsing was successfully completed")
            logger.info(f"Parsing time: {parsing_stop - parsing_start:.2f}")
            logger.info(f"Overall time: {parsing_stop - coro_start:.2f}")
            self._data = parsed[:]

    def copy(self) -> Any:
        """
        :return: copied object.
        """
        copy_obj = self.__class__(
            self.query, self.p_count, file=self.file, marker=self.marker, **self.params)
        copy_obj._data = self.data.copy()
        return copy_obj

    def sort_data(self,
                  **kwargs) -> None:
        """ Sort the data by using a key.

        :keyword key: func to sort, called to Example objects, by default – len.
        :keyword reverse: bool, whether the data will sort in reversed order,
         by default – False.
        :return None.
        :exception TypeError: if the key is uncallable.
        """
        key = kwargs.pop('key', lambda example: len(example))
        reverse = kwargs.pop('reverse', False)

        if not callable(key):
            logger.error("Given uncallable key to sort")
            raise TypeError("Sort key must be callable")
        self._data.sort(key=key, reverse=reverse)

    def pop(self,
            index: int) -> Any:
        """ Remove and return element from data at the index.

        :param index: int, index of the element.
        :return: Example object.
        """
        return self._data.pop(index)

    def shuffle(self) -> None:
        """ Shuffle list of examples.
        :return: None.
        """
        random.shuffle(self._data)

    def clear(self) -> None:
        """ Clear examples list.

        :return: None.
        """
        self._data.clear()

    def filter(self,
               key: Callable) -> None:
        """ Remove some items, that are not satisfied the key.

        :param key: callable, it will be used to Example
        objects inside the data list.
        :return: None.
        """
        filtered_data = list(filter(key, self.data))
        self._data = filtered_data[:]

    def findall(self,
                pattern: Any) -> None:
        """ Apply the pattern to the examples' text with re.findall.
        Change tha data list.

        :param pattern: r str or re.pattern, pattern to apply.
        :return: None.
        """
        pass

    def finditer(self,
                 pattern: Any) -> None:
        """ Apply the pattern to the examples' text with re.findall.
        Change tha data list.

        :param pattern: r str or re.pattern, pattern to apply.
        :return: None.
        """
        pass

    def __repr__(self) -> str:
        """ Format:
                Classname
                Length
                Database filename
                Request params
                Pages count
                Request

        :return: str with the format.
        """
        res = (f"{self.__class__.__name__}\n"
               f"{len(self)}\n" 
               f"{self.file}\n" 
               f"{self.params}\n" 
               f"{self.p_count}\n"
               f"{self.query}\n")
        return res

    def __str__(self) -> str:
        """
        :return: str, info about Corpus and enumerated examples.
        """
        q_forms = ', '.join(self.forms_in_query)
        metainfo = f"Russian National Corpus (https://ruscorpora.ru)\n" \
                   f"Class: {self.__class__.__name__}, len = {len(self)}\n" \
                   f"Pages: {self.p_count} of '{q_forms}' requested"

        data = self.data
        is_restricted = False
        if self.__RESTRICT_SHOW is not False and \
                len(data) > self.__RESTRICT_SHOW:
            data = self.data[:self.__RESTRICT_SHOW]
            is_restricted = True

        examples = '\n\n'.join(
            f"{num}.\n{str(example)}"
            for num, example in enumerate(data, 1)
        )
        if is_restricted:
            examples += '\n...'

        return f"{metainfo}\n\n{examples}"

    def __len__(self) -> int:
        """
        :return: int, count of examples.
        """
        return len(self.data)

    def __bool__(self) -> bool:
        """
        :return: bool, whether data exist.
        """
        return bool(self.data)

    def __call__(self) -> None:
        """ All the same to request_examples() """
        self.request_examples()

    def __iter__(self) -> iter:
        """
        :return: iter, iterator for data.
        """
        return iter(self.data)

    def __contains__(self,
                     item: Any) -> bool:
        """ Whether the Corpus obj contains the Example obj.

        :param item: obj with the same ex_type.
        :return: whether Corpus obj contains the Example obj.
        :exception TypeError: if wrong type (different Example) given.
        """
        if not isinstance(item, self.ex_type):
            msg = "in supports with the same Example objects"
            logger.error(msg)
            raise TypeError(msg)
        return any(
            item == example
            for example in self.data
        )

    def __getattr__(self,
                    item: str) -> str or int or List or None:
        """ Get request param.

        :param item: item, param name.
        :return: param value or None if it does not exist.
        """
        return self.params.get(item, None)

    def __getitem__(self,
                    item: int or slice) -> Any:
        """ Get example from data or create
        new obj with sliced data.

        :param item: int or slice.
        :return: one example or new obj with the same class and sliced data.
        :exception TypeError: if wrong type given.
        """
        if not isinstance(item, (int, slice)):
            logger.error("Int or slice expected")
            raise TypeError("Int or slice expected")

        if isinstance(item, int):
            return self.data[item]

        new_data = self.data[item]
        new_obj = self.copy()
        new_obj._data = new_data.copy()
        return new_obj

    def __setitem__(self,
                    key: int,
                    value: Any) -> None:
        """ Change the example.

        :param key: int, index of the example.
        :param value: ex_type, new example.
        :return: None.
        :exception TypeError: if wrong type given.
        """
        if not isinstance(value, self.ex_type):
            msg = f"Wrong type {type(value)}, " \
                  f"{type(self.ex_type)} expected"
            logger.error(msg)
            raise TypeError(msg)

        try:
            self._data[key] = value
        except Exception:
            logger.exception(f'Setting item: {value} to {key}')
            raise

    def __delitem__(self,
                    key: int or slice) -> None:
        """ Delete example at the index or
        remove several ones using slice.

        :param key: int or slice, address of item(s) to delete.
        :return: None.
        :exception: if sth went wrong.
        """
        try:
            del self._data[key]
        except Exception:
            logger.exception(f"Deleting item: {key}")
            raise

    def __copy__(self) -> Any:
        """ Copy self.

        :return: copied obj.
        """
        return self[:]


class MainCorpus(Corpus):
    _MODE = 'main'

    def __init__(self, *args, **kwargs):
        # for descendants
        ex_type = kwargs.pop('ex_type', expl.MainExample)
        super().__init__(*args, **kwargs, ex_type=ex_type)
        self._params['mode'] = self._MODE

    def _parse_example(self,
                       example: bs4.element.Tag):
        """ Parse example to Example object.

        :param example: tag, example to parse.
        :return: example obj.
        """
        src = Corpus._get_source(example)
        txt = Corpus._get_text(example)
        txt = txt[:txt.index(src)]
        txt = txt[:txt.rindex('[')].strip()

        doc_url = Corpus._get_doc_url(example)
        ambiguation = Corpus._get_ambiguation(example)
        found_words = Corpus._find_searched_words(example)

        new_ex = self.ex_type(txt, src, ambiguation, found_words, doc_url)
        new_ex.mark_found_words(self.marker)
        return new_ex

    def _parse_doc(self,
                   doc: bs4.element.Tag) -> List[expl.MainExample]:
        """ Parse document to list of examples.

        :param doc: bs4.element.ResultSet,
        """
        if not doc:
            logger.info(f"Empty doc found, params: {self.params}")
            return []
        res = []

        for example in doc.find_all('li'):
            new_ex = self._parse_example(example)
            res += [new_ex]
            self._add_wordforms(new_ex.found_wordforms)
        return res


class NGrams(Corpus):
    # env = sas1_2
    pass


class BiGrams(NGrams):
    pass


class ThreeGrams(NGrams):
    pass


class FourGrams(NGrams):
    pass


class FiveGrams(NGrams):
    pass


class SyntaxCorpus(Corpus):
    pass


class Paper2000Corpus(MainCorpus):
    _MODE = 'paper'

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs, ex_type=expl.Paper2000Example)
        self._params['mode'] = self._MODE


class PaperRegionalCorpus(MainCorpus):
    _MODE = 'regional'

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs, ex_type=expl.PaperRegionalExample)
        self._params['mode'] = self._MODE


class ParallelCorpus(Corpus):
    _MODE = 'para'

    def __init__(self, *args, **kwargs) -> None:
        # for descendants
        ex_type = kwargs.pop('ex_type', expl.ParallelExample)
        super().__init__(*args, **kwargs, ex_type=ex_type)
        self._params['mode'] = self._MODE

    def _parse_text(self,
                    lang: str,
                    text: bs4.element.Tag) -> Any:
        """ Parse one pair element of pair: original – translation.

        :param lang: str, language of text. This param
        :return: Example with one translation of text.
        ParallelExample supports '+=' method.
        """
        src = Corpus._get_source(text)
        ambiguation = Corpus._get_ambiguation(text)
        doc_url = Corpus._get_doc_url(text)
        txt = Corpus._get_text(text)
        # remove source from text
        txt = txt[:txt.index(src)]
        txt = txt[:txt.rindex('[')].strip()

        found_words = Corpus._find_searched_words(text)

        new_txt = self.ex_type(
            {lang: txt}, src, ambiguation, found_words, doc_url)
        new_txt.mark_found_words(self.marker)
        return new_txt

    def _parse_example(self,
                       tag: bs4.element.Tag) -> Any:
        """ Parse a pair: original – translation to Example.

        :param tag: bs4.element.Tag, pair to parse.
        :return: Example.
        """
        # this example's expected to have default args
        result_example = self.ex_type()

        langs = tag.find_all('td', {'class': "para-lang"})
        texts = tag.find_all('li')
        for lang, text in zip(langs, texts):
            lang = lang.text.strip()
            new_txt = self._parse_text(lang, text)
            result_example += new_txt
        return result_example

    def _parse_doc(self,
                   doc: bs4.element.Tag) -> List:
        """ Parse one document.

        :param doc: bs4.element.Tag, document to parse.
        :return: list of Examples.
        """
        res = []
        for example in doc.find_all('table', {'class': 'para'}):
            new_ex = self._parse_example(example)
            res += [new_ex]
            self._add_wordforms(new_ex.found_wordforms)
        return res

    def _load_data(self) -> List:
        """ Load data from csv file.

        :return: list of examples.
        """
        if self.out == 'kwic':
            return super()._load_data()
        with self.file.open('r', encoding='utf-16') as f:
            dm = self._DATA_W_DELIMITER
            qch = self._DATA_W_QUOTCHAR
            reader = csv.reader(f, delimiter=dm, quotechar=qch)

            columns = next(reader)
            end_lang_tags = columns.index('source')
            lang_tags = columns[:end_lang_tags]
            data = []

            for row in reader:
                # to create dict {lang: text in the lang}
                langs = {}
                for num, lang in enumerate(lang_tags):
                    langs[lang] = row[num]

                new_ex = self.ex_type(langs, *row[end_lang_tags:])
                data += [new_ex]

                self._add_wordforms(new_ex.found_wordforms)

        return data


class MultilingualParaCorpus(ParallelCorpus):
    _MODE = 'multi'

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs,
                         ex_type=expl.MultilingualParaExample)
        self._params['mode'] = self._MODE

    def _from_file(self) -> None:
        msg = "Working with files does not support"
        logger.error(msg)
        raise NotImplementedError(msg)

    def dump(self) -> None:
        msg = "Working with files does not support"
        logger.error(msg)
        raise NotImplementedError(msg)


class TutoringCorpus(MainCorpus):
    _MODE = 'school'

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs,
                         ex_type=expl.TutoringExample)
        self._params['mode'] = self._MODE


class DialectalCorpus(MainCorpus):
    _MODE = 'dialect'

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs,
                         ex_type=expl.DialectalExample)
        self._params['mode'] = self._MODE


# save lines
class PoeticCorpus(Corpus):
    pass


class SpokenCorpus(MainCorpus):
    _MODE = 'spoken'

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs,
                         ex_type=expl.SpokenExample)
        self._params['mode'] = self._MODE


class AccentologicalCorpus(MainCorpus):
    _MODE = 'accent'

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs,
                         ex_type=expl.AccentologicalExample)
        self._params['mode'] = self._MODE


class MultimodalCorpus(Corpus):
    MEDIA_FOLDER = Corpus.DATA_FOLDER / 'media'
    _MODE = 'murco'

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs, ex_type=expl.MultimodalExample)
        self._params['mode'] = self._MODE

    def _parse_example(self,
                       example: bs4.element.Tag) -> Tuple[str, str, str, list, str]:
        """ Parse example get text, source etc.

        :param example: bs4.element.Tag, example to parse.
        :return: tuple of parsed values.
        """
        src = Corpus._get_source(example)
        txt = Corpus._get_text(example)
        txt = txt[:txt.index(src)]
        txt = txt[:txt.rindex('[')].strip()

        doc_url = Corpus._get_doc_url(example)
        ambiguation = Corpus._get_ambiguation(example)
        found_words = Corpus._find_searched_words(example)

        return txt, src, ambiguation, found_words, doc_url

    def _parse_media(self,
                     media: bs4.element.Tag) -> Tuple[str, str]:
        """ Get link to the media file, filepath.

        :param media: bs4.element.Tag, here they are.
        :return: tuple of str and Path, link and filepath.
        """
        try:
            media_link = media.find('td').a['href']
        except Exception:
            raise
        media_link, filename = media_link.split('?name=')
        return media_link, self.MEDIA_FOLDER / filename

    def _parse_doc(self,
                   doc: bs4.element.Tag) -> List[Any]:
        """ Parse the documents to examples.

        :param doc: doc to parse.
        :return: list of examples.
        """
        try:
            media, example = doc.find_all('td', {'valign': 'top'})
        except ValueError:
            return []
        examples = []

        media_url, filename = self._parse_media(media)
        for example in example.find_all('li'):
            data_from_example = self._parse_example(example)

            new_ex = self.ex_type(*data_from_example, media_url, filename)
            new_ex.mark_found_words(self.marker)
            self._add_wordforms(new_ex.found_wordforms)
            examples += [new_ex]

        return examples

    def download_all(self) -> None:
        """ Download all files.

        :return: None.
        """
        os.makedirs(self.MEDIA_FOLDER, exist_ok=True)

        urls_to_names = [
            (example._media_url, example.filepath)
            for example in self
        ]
        creq.download_docs(urls_to_names)


class MultiPARCCorpus(Corpus):
    pass


class HistoricalCorpus(Corpus):
    pass



__all__ = (
    'MainExample',
    'SyntaxExample',
    'Paper2000Example',
    'PaperRegionalExample',
    'ParallelExample',
    'MultilingualParaExample',
    'TutoringExample',
    'DialectalExample',
    'PoeticExample',
    'SpokenExample',
    'AccentologicalExample',
    'MultimodalExample',
    'MultiPARCExample',
    'HistoricalExample',
    'TutoringExample',
    'KwicExample'
)

import logging
import os
import re
import webbrowser
from pathlib import Path
from typing import List, Callable, Dict, Any

import rnc.corpora_requests as creq

logger = logging.getLogger("rnc")


def mark_found_words(txt: str,
                     words: List[str],
                     marker: Callable) -> str:
    """ Mark words in the text.

    :param txt: str, text.
    :param words: list of str, words to mark.
    :param marker: callable, function to mark words with it.
    :return: str with marked words.
    """
    if marker is None:
        return txt

    for word in words:
        txt = re.sub(fr'\b{word}\b', marker(word), txt)
    return txt


# TODO
class TextInfo:
    pass


class Example:
    """ Base examples class """
    __slots__ = (
        '_txt', '_src', '_doc_url',
        '_ambiguation', '_found_wordforms')

    def __init__(self,
                 txt: str,
                 src: str,
                 ambiguation: str,
                 found_wordforms: List[str] or str,
                 doc_url: str) -> None:
        """
        :param txt: str, example's text.
        :param src: str, example's source.
        :param ambiguation: str, example's ambiguation.
        :param found_wordforms: list of str, None or str joined with ', ',
        example's found wordforms.
        :param doc_url: str, example's url.
        :return: None.
        """
        self._txt = txt
        self._src = src
        self._ambiguation = ambiguation
        self._doc_url = doc_url

        wf = found_wordforms or []
        if isinstance(wf, str):
            wf = found_wordforms.split(', ')
        self._found_wordforms = wf

    @property
    def txt(self) -> Any:
        """
        :return: any type, example's text.
        """
        return self._txt

    @property
    def src(self) -> Any:
        """
        :return: any type, example's source.
        """
        return self._src

    @property
    def ambiguation(self) -> Any:
        """
        :return: any type, example's ambiguation.
        """
        return self._ambiguation

    @property
    def doc_url(self) -> str:
        """
        :return: str, example's URL.
        """
        return self._doc_url

    @property
    def found_wordforms(self) -> List[str]:
        """
        :return: list of str, example's found wordforms.
        """
        return self._found_wordforms

    @property
    def columns(self) -> List[str]:
        """ For csv writing.

        :return: list of str, names of columns.
        """
        return list(self.data.keys()) + ['URL']

    @property
    def items(self) -> List[Any]:
        """ For csv writing.

        :return: list of any types, values of columns.
        """
        # ATTENTION:
        # these order must be the same as in the constructor
        return list(self.data.values()) + [self.doc_url]

    @property
    def data(self) -> Dict[str, Any]:
        """ There are all fields except for doc_url.
        Found wordforms joined with ', '.

        :return: dict with fields names and their values.
        """
        data = {
            'text': self.txt,
            'source': self.src,
            'ambiguation': self.ambiguation,
            'found wordforms': ', '.join(self.found_wordforms)
        }
        return data

    @txt.setter
    def txt(self,
            other: Any) -> None:
        """ Set text.

        :param other: any type, new txt.
        :return: None.
        """
        if not isinstance(other, str):
            class_name = self.__class__.__name__
            logger.warning(f"As a text to {class_name} "
                           f"set {type(other)}, str expected")

        self._txt = other

    @src.setter
    def src(self,
            other: Any) -> None:
        """ Set source.

        :param other: any type, new src.
        :return: None.
        """
        if not isinstance(other, str):
            class_name = self.__class__.__name__
            logger.warning(f"As a source to {class_name} "
                           f"set {type(other)}, str expected")

        self._src = other

    @ambiguation.setter
    def ambiguation(self,
                    other: Any) -> None:
        """ Set ambiguation.

        :param other: any type, new ambiguation.
        :return: None.
        """
        if not isinstance(other, str):
            class_name = self.__class__.__name__
            logger.warning(f"As a ambiguation to {class_name} "
                           f"set {type(other)}, str expected")
        self._ambiguation = other

    def open_doc(self) -> None:
        """ Open the doc in the new tab of the default browser.

        :return: None.
        :exception: if something is wrong.
        """
        try:
            webbrowser.open_new_tab(self._doc_url)
        except Exception:
            logger.exception(
                f"Error while opening doc with url: {self._doc_url}")
            raise

    def mark_found_words(self,
                         marker: Callable) -> None:
        """ Mark found wordforms in the text with marker.

        :param marker: function to mark found wordforms.
        :return: None.
        """
        self._txt = mark_found_words(
            self.txt, self.found_wordforms, marker)

    def copy(self) -> Any:
        """
        :return: copied obj.
        """
        return self.__class__(*self.data.values(), self.doc_url)

    def __eq__(self,
               other: Any) -> bool:
        """ ==

        :param other: other Example object.
        :return: bool, whether data equal.
        """
        return self.items == other.items

    def __contains__(self,
                     item: Any) -> bool:
        """ Whether the item is in the text.
        Registers equaled (if it is available).

        :param item: any type, item to check.
        :return: whether item is in text.
        """
        try:
            return item.lower() in self.txt.lower()
        except AttributeError:
            return item in self.txt

    def __str__(self) -> str:
        """ Str format:
            TEXT: ...
            SOURCE: ...
            AMBIGUATION: ...
            FOUND WORDFORMS: ...

        :return: this str.
        """
        res = '\n'.join(
            f"{key.upper()}: {val}"
            for key, val in self.data.items()
        )
        return res

    def __repr__(self) -> str:
        """ Str format:
            text: ...
            source: ...
            ambiguation: ...
            found wordforms: ...
            URL: ...

        :return: this str.
        """
        fields = '\n'.join(
            f"{key}: {val}"
            for key, val in self.data.items()
        )
        url = f"URL: {self.doc_url}"
        return f"{fields}\n{url}"

    def __hash__(self) -> int:
        """ Hash str with all example fields.

        :return: int, hash.
        """
        return hash(repr(self))

    def __bool__(self) -> bool:
        """ .
        :return: bool, whether fields (expect for url) exist
        """
        return all(val for val in self.data.values())


class KwicExample(Example):
    __slots__ = '_left', '_center', '_right'

    def __init__(self,
                 left: str,
                 center: str,
                 right: str,
                 src: str,
                 found_wordforms: List[str] or str,
                 doc_url: str) -> None:
        """ There is no ambiguation, it set with ''.

        :param left: str, example's left context.
        :param center: str, example's center context.
        :param right: str, example's right context.
        :param src: str, example's source.
        :param found_wordforms: list of str or str, example's found wordforms.
        :param doc_url: str, example's URL.
        """
        self._left = left
        self._center = center
        self._right = right
        super().__init__(self.txt, src, '', found_wordforms, doc_url)

    @property
    def left(self) -> Any:
        """
        :return: str, example's left context.
        """
        return self._left

    @property
    def center(self) -> Any:
        """
        :return: str, example's center context.
        """
        return self._center

    @property
    def right(self) -> Any:
        """
        :return: str, example's right context.
        """
        return self._right

    @property
    def txt(self) -> str:
        """
        :return: str, joined left, center and right contexts.
        """
        # it's assumed that all contexts are stripped
        return f"{self.left} {self.center} {self.right}"

    @property
    def ambiguation(self) -> None:
        msg = "There is no ambiguation in KWICExamples"
        logger.error(msg)
        raise NotImplementedError(msg)

    @property
    def data(self) -> Dict[str, Any]:
        """ All fields except for URL.

        :return: dict with fields' names and their values.
        """
        data = {
            'left': self.left,
            'center': self.center,
            'right': self.right,
            'source': self.src,
            'found wordforms': ', '.join(self.found_wordforms)
        }
        return data

    @left.setter
    def left(self,
             other: Any) -> None:
        """ Set left context.

        :param other: any type, new left context.
        :return: None.
        """
        if not isinstance(other, str):
            class_name = self.__class__.__name__
            logger.warning(f"As a left context to {class_name} "
                           f"set {type(other)}, str expected")
        self._left = other

    @center.setter
    def center(self,
               other: Any) -> None:
        """ Set center context.

        :param other: any type, new center context.
        :return: None.
        """
        if not isinstance(other, str):
            class_name = self.__class__.__name__
            logger.warning(f"As a center context to {class_name} "
                           f"set {type(other)}, str expected")
        self._center = other

    @right.setter
    def right(self,
              other: Any) -> None:
        """ Set right context.

        :param other: any type, new right context.
        :return: None.
        """
        if not isinstance(other, str):
            class_name = self.__class__.__name__
            logger.warning(f"As a right context to {class_name} "
                           f"set {type(other)}, str expected")
        self._right = other

    @txt.setter
    def txt(self,
            other: Any) -> None:
        """ Text setter not implemented to KWICExamples """
        msg = "Text setter not implemented to KWICExamples"
        logger.error(msg)
        raise NotImplementedError(msg)

    def mark_found_words(self,
                         marker: Callable) -> None:
        """ Mark found wordforms in all contexts using marker.

        :param marker: function to mark found wordforms.
        :return: None.
        """
        words = self.found_wordforms
        self._left = mark_found_words(self.left, words, marker)
        self._center = mark_found_words(self.center, words, marker)
        self._right = mark_found_words(self.right, words, marker)


class MainExample(Example):
    pass


class SyntaxExample(Example):
    pass


class Paper2000Example(Example):
    pass


class PaperRegionalExample(Example):
    pass


class ParallelExample(Example):
    def __init__(self,
                 txt: Dict[str, str] = None,
                 src: str = '',
                 ambiguation: str = '',
                 found_wordforms: List[str] or str = None,
                 doc_url: str = '') -> None:
        """
        :param txt: dict of str, {language tag: text}
        :param src: str, examples source.
        :param ambiguation: str, examples ambiguation.
        :param found_wordforms: list of str, examples found wordforms.
        :param doc_url: str, examples URL.
        """
        txt = txt or {}

        super().__init__(txt, src, ambiguation, found_wordforms, doc_url)
        self.sort()

    @property
    def txt(self) -> Dict[str, Any]:
        """ Get dict with texts.

        :return: dict of any types.
        """
        return self._txt

    @txt.setter
    def txt(self,
            other: Any) -> None:
        """ Text setter not implemented to the ParallelExample """
        msg = ("Text setter not implemented to ParallelExamples. "
               "Try to use ex['language tag'] instead")
        logger.error(msg)
        raise NotImplementedError(msg)

    @property
    def data(self) -> Dict[str, Any]:
        """ There are all fields except for doc_url.
        Found wordforms joined with ', '.

        :return: dict with fields' names and their values.
        """
        data = self.txt.copy()
        data['source'] = self.src
        data['ambiguation'] = self.ambiguation
        data['found wordforms'] = ', '.join(self.found_wordforms)

        return data

    def mark_found_words(self,
                         marker: Callable) -> None:
        """ Mark found wordforms in the text with marker.

        :param marker: function to mark.
        :return: None.
        """
        for lang, txt in self.txt.items():
            self[lang] = mark_found_words(
                txt, self.found_wordforms, marker)

    @staticmethod
    def _best_src(f_src: str,
                  s_src: str) -> str:
        """ Choose the best source, means
        there are two translations in it.

        :param f_src: str, first source.
        :param s_src: str, second source.
        :return: str, best of them.
        """
        if '|' in s_src:
            return s_src
        return f_src

    def sort(self,
             key: Callable = None,
             reverse: bool = False) -> None:
        """ Sort txt dict, allowing the key to
        items() from there.

        :param key: callable, key to sort. Sort by language tag by default.
        :param reverse: bool, whether sorting will be in reversed order.
        :return: None.
        """
        key = key or (lambda items: items[0])
        data = sorted(self.txt.items(), key=key, reverse=reverse)
        self._txt = dict(data)

    def copy(self) -> Any:
        """
        :return: copied obj.
        """
        txt_copy = self.txt.copy()
        return self.__class__(
            txt_copy, self.src, self.ambiguation,
            self.found_wordforms, self.doc_url
        )

    def __contains__(self,
                     item: Any) -> bool:
        """ Whether the item is in the text.
        Registers equaled (if it is available).

        :param item: any type, item to check.
        :return: whether item is in text.
        """
        try:
            return any(item.lower() in i.lower() for i in self.txt.values())
        except AttributeError:
            return any(item in i for i in self.txt.values())

    def __iadd__(self,
                 other: Any) -> Any:
        """ Concatenate two examples:
            – join the texts

            – choose the best source, there are
            two translations there

            – choose where the text is disambiguated.

            – join found wordforms.

        :param other: instance of the same class.
        :return: self.
        :exception TypeError: if wrong type given.
        """
        if not isinstance(other, self.__class__):
            class_name = self.__class__.__name__
            msg = (f"'+=' supported only with '{class_name}' "
                   f"objects, but {type(other)} given")
            logger.error(msg)
            raise TypeError(msg)

        for lang, txt in other.txt.items():
            new_txt = f"{self.txt.get(lang, '')} {txt}"
            self[lang] = new_txt.lstrip()
        # save the order of languages
        self.sort(lambda x: x[0])
        # source contains two translations
        self._src = self.__class__._best_src(self.src, other.src)

        # for MultilingualParaCorpus
        if not self.src:
            self._src = other.src

        o_amb = other.ambiguation
        if not self.ambiguation or o_amb and 'not' not in o_amb:
            self._ambiguation = o_amb
        if not self.doc_url:
            self._doc_url = other.doc_url
        self._found_wordforms += other.found_wordforms

        return self

    def __getattr__(self,
                    item: str) -> Any:
        """ Get the text in language.

        :param item: str, language tag.
        :return: str or None, text in the language if exists.
        """
        try:
            return getattr(super(), item)
        except AttributeError:
            return self.txt.get(item, None)

    def __getitem__(self,
                    lang: str) -> Any:
        """ Get text in the language.

        :param lang: str, language tag.
        :return: str or None, text in the language if exists.
        """
        return self.txt.get(lang, None)

    def __setitem__(self,
                    lang: str,
                    txt: Any) -> None:
        """ Change text in the language.

        :param lang: str, language tag.
        :param txt: any type, new text.
        :return: None.
        :exception ValueError: if the text in the language does not exist.
        """
        if not isinstance(txt, str):
            class_name = self.__class__.__name__
            logger.warning(f"As a '{lang}' to {class_name} "
                           f"set {type(txt)}, str expected")
        self._txt[lang] = txt


class MultilingualParaExample(ParallelExample):
    pass


class TutoringExample(Example):
    pass


class DialectalExample(Example):
    pass


class PoeticExample(Example):
    pass


class SpokenExample(Example):
    pass


class AccentologicalExample(Example):
    pass


class MultimodalExample(Example):
    def __init__(self,
                 txt: str,
                 src: str,
                 ambiguation: str,
                 found_wordforms: List[str] or str,
                 doc_url: str,
                 media_url: str,
                 filename: str) -> None:
        """
        :param txt: str, example's text.
        :param src: str, example's source.
        :param ambiguation: str, example's ambiguation.
        :param found_wordforms: list of str, None or str joined with ', ',
        example's found wordforms.
        :param doc_url: str, example's url.
        :param media_url: str, URL to example's media file.
        :param filename: str or Path, default name of the file (from RNC).
        :return: None.
        """
        super().__init__(txt, src, ambiguation, found_wordforms, doc_url)
        self._media_url = media_url
        self._filepath = Path(filename)

    @property
    def filepath(self) -> Path:
        """ Get the path to the local file.

        :return: Path to the file.
        """
        return self._filepath

    @filepath.setter
    def filepath(self,
                 other: str or Path) -> None:
        """ Set new path to the local file.

        ATTENTION: if the file exists it will not be moved to
        the new path. You should call 'download_file()' again.

        :param other: str or Path, new path to the local file.
        :return: None.
        """
        self._filepath = Path(other)

    @property
    def columns(self) -> List[str]:
        """ For csv writing.

        :return: list of str, names of columns.
        """
        return super().columns + ['media_url', 'filename']

    @property
    def items(self) -> List[Any]:
        """ For csv writing.

        :return: list of any types, values of columns.
        """
        return super().items + [self._media_url, self.filepath]

    def download_file(self) -> None:
        """ Download the media file.

        :return: None.
        """
        os.makedirs(self.filepath.parent, exist_ok=True)

        data = [(self._media_url, str(self.filepath))]
        try:
            creq.download_docs(data)
        except Exception:
            raise


class MultiPARCExample(Example):
    pass


class HistoricalExample(Example):
    pass



__version__ = '0.6'

import logging
import os
from pathlib import Path

DEFAULT_MSG_FMT = "[{name}:{module}:{levelname}:{funcName}:{asctime}] {message}"
DEFAULT_DATE_FMT = "%d.%m.%Y %H:%M:%S"

LOG_FOLDER = Path('logs')
os.makedirs(LOG_FOLDER, exist_ok=True)


def create_formatter(message_format: str = DEFAULT_MSG_FMT,
                     date_format: str = DEFAULT_DATE_FMT,
                     style: str = '{') -> logging.Formatter:
    """ Create message formatter.

    :param message_format: str, way to format message.
    :param date_format: str, way to format date.
     Str with %, because it is needed for time.strftime()
    :param style: str, % or {. Whether message_format
     contains % or { way to format.
    :return: logging.Formatter.
    """
    formatter = logging.Formatter(
        fmt=message_format,
        datefmt=date_format,
        style=style
    )
    return formatter


def create_stream_handler(level=logging.NOTSET,
                          formatter: logging.Formatter = None) -> logging.StreamHandler:
    """ Create stream handler.

    :param level: handler level. WARNING by default.
    :param formatter: Formatter, message format.
    :return: stream handler.
    """
    stream_handler = logging.StreamHandler()
    stream_handler.setLevel(level)
    formatter = formatter or create_formatter()
    stream_handler.setFormatter(formatter)

    return stream_handler


def create_file_handler(level=logging.DEBUG,
                        log_path: str = None,
                        formatter: logging.Formatter = None,
                        **kwargs) -> logging.FileHandler:
    """ Create file handler.

    :param level: handler level.
    :param log_path: str, path to log file.
    :param formatter: Formatter, message format.
    :keyword delay: bool, whether the file will not be opened
     until the first logger calling. True by default.
    :return: file handler.
    """
    delay = kwargs.pop('delay', True)

    file_handler = logging.FileHandler(
        log_path, delay=delay, encoding='utf-8', **kwargs)
    file_handler.setLevel(level)
    formatter = formatter or create_formatter()
    file_handler.setFormatter(formatter)

    return file_handler


def create_logger(module_name: str,
                  level=logging.DEBUG,
                  *handlers) -> logging.Logger:
    """ Create logger.

    :param module_name: str, name of logger.
    :param level: level of logger, DEBUG by default.
    :param handlers: handlers to be added to the logger.
    :return: logger.
    """
    logger_ = logging.getLogger(module_name)
    logger_.setLevel(level)

    path = LOG_FOLDER / f"{module_name}.log"

    handlers = handlers or [
        create_stream_handler(),
        create_file_handler(log_path=path)
    ]

    for handler in handlers:
        logger_.addHandler(handler)

    return logger_


logger = create_logger("rnc")

from .corpora import (
    MainCorpus,
    Paper2000Corpus,
    PaperRegionalCorpus,
    ParallelCorpus,
    DialectalCorpus,
    SpokenCorpus,
    AccentologicalCorpus,
    MultilingualParaCorpus,
    TutoringCorpus,
    MultimodalCorpus
)
from .corpora_params import Subcorpus
from .examples import (
    MainExample,
    Paper2000Example,
    PaperRegionalExample,
    ParallelExample,
    DialectalExample,
    SpokenExample,
    AccentologicalExample,
    MultilingualParaExample,
    TutoringExample,
    MultimodalExample,
    KwicExample
)


def set_handler_level(level, handler_class) -> None:
    for handler_index in range(len(logger.handlers)):
        if logger.handlers[handler_index].__class__ == handler_class:
            logger.handlers[handler_index].setLevel(level)
            return
    raise ValueError(f"There is no '{handler_class}' handler")


def set_stream_handler_level(level):
    try:
        set_handler_level(level, logging.StreamHandler)
    except ValueError:
        raise


def set_file_handler_level(level):
    try:
        set_handler_level(level, logging.FileHandler)
    except ValueError:
        raise


def set_logger_level(level):
    logger.setLevel(level)


subcorpus = Subcorpus()

__all__ = (
    'MainCorpus',
    'Paper2000Corpus',
    'PaperRegionalCorpus',
    'ParallelCorpus',
    'DialectalCorpus',
    'SpokenCorpus',
    'AccentologicalCorpus',
    'MultilingualParaCorpus',
    'TutoringCorpus',
    'MultimodalCorpus',
    'subcorpus',

    'MainExample',
    'Paper2000Example',
    'PaperRegionalExample',
    'ParallelExample',
    'DialectalExample',
    'SpokenExample',
    'AccentologicalExample',
    'MultilingualParaExample',
    'TutoringExample',
    'MultimodalExample',
    'KwicExample',

    'set_stream_handler_level',
    'set_file_handler_level',
    'set_logger_level'
)



import os
from pathlib import Path

import rnc.corpora as corp
import rnc.examples as expl
import pytest


class TemplateTestExamples:
    ex = expl.Example

    def test_text_getter(self):
        assert isinstance(self.ex.txt, str) and self.ex.txt

    def test_text_setter(self):
        copy = self.ex.copy()
        new_txt = 'test_text'
        copy.txt = new_txt

        assert copy.txt == new_txt

    def test_text_setter_with_notstr_type(self):
        copy = self.ex.copy()
        new_txt = list('new ex')
        copy.txt = new_txt

        assert copy.txt == new_txt

    def test_src_getter(self):
        assert isinstance(self.ex.src, str) and self.ex.src

    def test_src_setter(self):
        copy = self.ex.copy()
        new_src = 'test_source'
        copy.src = new_src

        assert copy.src == new_src

    def test_src_setter_with_notstr_type(self):
        copy = self.ex.copy()
        new_src = list('test_source')
        copy.src = new_src

        assert copy.src == new_src

    def test_ambiguation_getter(self):
        amb = self.ex.ambiguation

        assert isinstance(amb, str) and amb

    def test_ambiguation_setter(self):
        copy = self.ex.copy()
        new_amb = 'new_amb'
        copy.ambiguation = new_amb

        assert copy.ambiguation == new_amb

    def test_ambiguation_setter_with_nostr_type(self):
        copy = self.ex.copy()
        new_amb = list('new_amb')
        copy.ambiguation = new_amb

        assert copy.ambiguation == new_amb

    def test_doc_url(self):
        url = self.ex.doc_url

        assert isinstance(url, str) and url

    def test_found_worforms(self):
        wf = self.ex.found_wordforms

        assert isinstance(wf, list) and wf

    def test_columns(self):
        columns = self.ex.columns
        expected = ['text', 'source', 'ambiguation', 'found wordforms', 'URL']

        assert columns == expected

    def test_items(self):
        items = self.ex.items

        assert all(isinstance(i, str) and i for i in items)

    def test_data(self):
        data = self.ex.data

        assert isinstance(data, dict) and len(data) is 4

    def test_open_doc(self):
        self.ex.open_doc()

    def test_mark_words(self):
        txt_template = "{}, aghA{}12g_HGH g,.,,{},.,a{}s agsd :'{}'!"
        copy = self.ex.copy()
        word = copy.found_wordforms[0]

        new_txt = txt_template.format(*([word] * 5))
        expected_txt = txt_template.format(
            word.upper(), word, word.upper(), word, word.upper())

        copy.txt = new_txt
        copy.mark_found_words(str.upper)

        assert copy.txt == expected_txt

    def test_deleter(self):
        with pytest.raises(AttributeError):
            del self.ex.txt

    def test_copy(self):
        assert self.ex == self.ex.copy()

    def test_contains(self):
        item = self.ex.txt[:5]

        assert item in self.ex

    def test_contains_with_notstr_type(self):
        copy = self.ex.copy()
        copy.txt = ['1', 2, 3, 4, '5']

        assert '1' in copy and 2 in copy

    def test_equal(self):
        lhs = rhs = self.ex.copy()
        lhs.txt = list('test lhs')
        rhs.txt = lhs.txt

        assert lhs == rhs


class TestMainExample(TemplateTestExamples):
    corpus_res = corp.MainCorpus('тест', 1, marker=str.upper)
    corpus_res.request_examples()
    ex: expl.MainExample = corpus_res[0]


class TestKwicExample(TemplateTestExamples):
    corpus_res = corp.MainCorpus('тест', 1, marker=str.upper, out='kwic')
    corpus_res.request_examples()
    ex: expl.KwicExample = corpus_res[0]

    def test_text_setter(self):
        with pytest.raises(NotImplementedError):
            self.ex.txt = 'sth'

    def test_text_setter_with_notstr_type(self):
        with pytest.raises(NotImplementedError):
            self.ex.txt = 123

    def test_left_setter(self):
        copy = self.ex.copy()
        new_left = 'new left'
        copy.left = new_left

        assert copy.left == new_left

    def test_left_setter_with_notstr_type(self):
        copy = self.ex.copy()
        new_left = list('new left')
        copy.left = new_left

        assert copy.left == new_left

    def test_center_setter(self):
        copy = self.ex.copy()
        new_center = 'new center'
        copy.center = new_center

        assert copy.center == new_center

    def test_center_setter_with_notstr_type(self):
        copy = self.ex.copy()
        new_center = list('new center')
        copy.center = new_center

        assert copy.center == new_center

    def test_right_setter(self):
        copy = self.ex.copy()
        new_right = 'new right'
        copy.right = new_right

        assert copy.right == new_right

    def test_right_setter_with_notstr_type(self):
        copy = self.ex.copy()
        new_right = list('new right')
        copy.right = new_right

        assert copy.right == new_right

    def test_ambiguation_getter(self):
        with pytest.raises(NotImplementedError):
            amb = self.ex.ambiguation

    def test_ambiguation_setter(self):
        with pytest.raises(AttributeError):
            self.ex.ambiguation = 'sth'

    def test_ambiguation_setter_with_nostr_type(self):
        with pytest.raises(AttributeError):
            self.ex.ambiguation = 123

    def test_columns(self):
        columns = self.ex.columns
        expected = ['left', 'center', 'right', 'source',
                    'found wordforms', 'URL']

        assert columns == expected

    def test_data(self):
        data = self.ex.data

        assert isinstance(data, dict) and len(data) is 5

    def test_mark_words(self):
        txt_template = "{}, aghA{}12g_HGH g,.,,{},.,a{}s agsd :'{}'!"
        copy = self.ex.copy()
        word = copy.found_wordforms[0]

        new_txt = txt_template.format(*([word] * 5))
        expected_txt = txt_template.format(
            word.upper(), word, word.upper(), word, word.upper())

        copy.left = copy.center = copy.right = new_txt
        copy.mark_found_words(str.upper)

        assert copy.left == copy.center == copy.right == expected_txt

    def test_contains_with_notstr_type(self):
        copy = self.ex.copy()
        copy.left = [1, 2, 3]
        copy.center = '1'
        copy.right = 12

        assert '12' in copy

    def test_equal(self):
        lhs = rhs = self.ex.copy()
        lhs.left = rhs.left = ''

        assert lhs == rhs


class TestParallelExample(TemplateTestExamples):
    corpus_res = corp.ParallelCorpus(
        'тест', 1, marker=str.upper, subcorpus=corp.Subcorpus.Parallel.English)
    corpus_res.request_examples()
    ex: expl.ParallelExample = corpus_res[0]

    def test_text_getter(self):
        txt = self.ex.txt

        assert isinstance(txt, dict) and len(txt) is 2

    def test_text_setter(self):
        with pytest.raises(NotImplementedError):
            self.ex.txt = {}

    def test_text_setter_with_notstr_type(self):
        with pytest.raises(NotImplementedError):
            self.ex.txt = 12

    def test_data(self):
        data = self.ex.data

        assert isinstance(data, dict) and len(data) is 5

    def test_iadd_with_nochanges(self):
        copy = self.ex.copy()
        new_ex = expl.ParallelExample()
        copy += new_ex

        assert copy == self.ex

    def test_iadd_with_extend_texts(self):
        txt = {
            'ru': 'текст1',
            'en': 'text1'
        }
        lhs = expl.ParallelExample(txt, 'old_src', 'not amb', ['test'])

        update = {
            'ru': 'и текст2',
            'en': 'and text2'
        }
        rhs = expl.ParallelExample(update, 'old src | new src', 'amb', ['test1'])
        lhs += rhs

        assert lhs.ru == 'текст1 и текст2'
        assert lhs.en == 'text1 and text2'
        assert lhs.src == 'old src | new src'
        assert lhs.ambiguation == 'amb'
        assert lhs.found_wordforms == ['test', 'test1']

    def test_iadd_with_new_lang(self):
        txt = {
            'ru': 'текст1',
            'en': 'text1'
        }
        lhs = expl.ParallelExample(txt, 'old src | new src', 'amb', ['test'])

        update = {
            'fr': 'и текст2',
            'ch': 'and text2'
        }
        rhs = expl.ParallelExample(update, 'new src', 'not amb', [])
        lhs += rhs

        assert lhs.ru == 'текст1'
        assert lhs.en == 'text1'
        assert lhs.fr == 'и текст2'
        assert lhs.ch == 'and text2'
        assert lhs.src == 'old src | new src'
        assert lhs.ambiguation == 'amb'
        assert lhs.found_wordforms == ['test']

    def test_iadd_with_new_lang_and_extend_texts(self):
        txt = {
            'ru': 'текст1',
            'en': 'text1'
        }
        lhs = expl.ParallelExample(txt, 'old src', 'not amb', ['test'])

        update = {
            'ru': 'и текст2',
            'en': 'and text2',
            'fr': 'и текст3',
            'ch': 'and text3'
        }
        rhs = expl.ParallelExample(update, '', '', ['text1'])
        lhs += rhs

        assert lhs.ru == 'текст1 и текст2'
        assert lhs.en == 'text1 and text2'
        assert lhs.fr == 'и текст3'
        assert lhs.ch == 'and text3'
        assert lhs.src == 'old src'
        assert lhs.ambiguation == 'not amb'
        assert lhs.found_wordforms == ['test', 'text1']

    def test_iadd_with_update_ambiguation_from_null(self):
        txt = {
            'ru': 'текст1',
            'en': 'text1'
        }
        lhs = expl.ParallelExample(txt, 'old src', '', [])

        update = {
            'fr': 'и текст2',
            'ch': 'and text2'
        }
        rhs = expl.ParallelExample(update, 'new src', 'not amb', [])
        lhs += rhs

        assert lhs.ru == 'текст1'
        assert lhs.en == 'text1'
        assert lhs.fr == 'и текст2'
        assert lhs.ch == 'and text2'
        assert lhs.src == 'old src'
        assert lhs.ambiguation == 'not amb'
        assert lhs.found_wordforms == []

    def test_columns(self):
        columns = self.ex.columns
        expected = ['en', 'ru', 'source', 'ambiguation',
                    'found wordforms', 'URL']

        assert columns == expected

    def test_getattr(self):
        txt = self.ex.ru

        assert isinstance(txt, str) and txt

    def test_getattr_none(self):
        assert self.ex.fr is None

    def test_getitem(self):
        txt = self.ex['ru']

        assert isinstance(txt, str) and txt

    def test_getitem_none(self):
        assert self.ex['fr'] is None

    def test_setitem(self):
        new_txt = 'sth with new txt'
        copy = self.ex.copy()
        copy['ru'] = new_txt

        assert copy.ru == new_txt

    def test_setitem_with_notstr_type(self):
        copy = self.ex.copy()
        new_txt = list('new txt')
        copy['ru'] = new_txt

        assert copy.ru == new_txt

    def test_sort(self):
        copy = self.ex.copy()
        copy.sort()

        assert copy.txt == self.ex.txt

    def test_mark_words(self):
        txt_template = "{}, aghA{}12g_HGH g,.,,{},.,a{}s agsd :'{}'!"
        copy = self.ex.copy()
        word = copy.found_wordforms[0]

        new_txt = txt_template.format(*([word] * 5))
        expected_txt = txt_template.format(
            word.upper(), word, word.upper(), word, word.upper())

        copy['ru'] = new_txt
        copy.mark_found_words(str.upper)

        assert copy.ru == expected_txt

    def test_equality(self):
        assert self.ex['ru'] == self.ex.ru

    def test_equal(self):
        lhs = rhs = self.ex.copy()
        lhs['fr'] = rhs['fr'] = ''

        assert lhs == rhs

    def test_contains(self):
        text = self.ex.ru[:5]

        assert text in self.ex

    def test_contains_with_notstr_type(self):
        copy = self.ex.copy()
        new_txt = ['1', 2, '3', '5']
        copy['ru'] = new_txt

        assert '1' in copy and '3' in copy


class TestMultimodalExample(TemplateTestExamples):
    corpus_res = corp.MultimodalCorpus(
        'корпус', 1, marker=str.upper)
    corpus_res.request_examples()
    ex: expl.MultimodalExample = corpus_res[0]

    def test_columns(self):
        columns = self.ex.columns
        expected = ['text', 'source', 'ambiguation',
                    'found wordforms', 'URL', 'media_url', 'filename']

        assert columns == expected

    def test_items(self):
        items = self.ex.items

        assert all(isinstance(i, (str, Path)) and i for i in items)

    def test_filepath_getter(self):
        path = self.ex.filepath

        assert isinstance(path, Path)

    def test_filepath_setter(self):
        self.ex.filepath = 'path'

    def test_download_file(self):
        self.ex.download_file()
        files = os.listdir('data\\media')

        assert self.ex.filepath.name in files



"""
Module for requesting to URL and get page's html code from there,
download media files, check that the request if correct, page in RNC exists.
"""

__all__ = 'get_htmls', 'is_request_correct'

import asyncio
import logging
from typing import List, Tuple

import aiofiles
import aiohttp
import aiojobs
import bs4

logger = logging.getLogger("rnc")
WAIT = 24


async def fetch(url: str,
                ses: aiohttp.ClientSession,
                queue: asyncio.Queue,
                **kwargs) -> None:
    """ Coro, obtaining page's html code.

    If response status == 429 sleep and try again.

    :param url: str, URL to get its html code.
    :param ses: aiohttp.ClientSession.
    :param queue: asyncio.Queue, queue to where put the results:
    'p' key to sort results and html code, decoded to UTF-8.
    :param kwargs: HTTP tags.
    :return: None.
    :exception: all exceptions should be processed here.
    """
    try:
        resp = await ses.get(url, params=kwargs)
    except Exception:
        logger.exception("Cannot get answer from RNC")
        return

    try:
        if resp.status != 429:
            resp.raise_for_status()
    except Exception:
        logger.error(
            f"{resp.status}: {resp.reason} requesting to {resp.url}"
        )
        resp.close()
        return

    if resp.status == 200:
        text = await resp.text('utf-8')
        await queue.put((kwargs['p'], text))
        resp.close()
    elif resp.status == 429:
        logger.debug(
            f"429, {resp.reason} Page: {kwargs['p']}, wait {WAIT}s"
        )
        resp.close()
        await asyncio.sleep(WAIT)
        await fetch(url, ses, queue, **kwargs)
    else:
        logger.critical("This case must not happened")


async def get_htmls_coro(url: str,
                         start: int,
                         stop: int,
                         **kwargs) -> List[str]:
    """ Coro doing requests and catching exceptions.

    URLs will be created for i in range(start, stop),
    HTTP tag 'p' (page) is i.

    :param url: str, URL.
    :param start: int, start page index.
    :param stop: int, stop page index.
    :param kwargs: HTTP tags.
    :return: list of str, html codes of the pages.
    """
    timeout = aiohttp.ClientTimeout(WAIT)
    # this limit might be wrong, it is
    # just not demanded to set any limit
    queue = asyncio.Queue(maxsize=100_000)

    async with aiohttp.ClientSession(timeout=timeout) as ses:
        scheduler = await aiojobs.create_scheduler(limit=None)
        for p_index in range(start, stop):
            await scheduler.spawn(
                fetch(url, ses, queue, **kwargs, p=p_index)
            )

        while len(scheduler) is not 0:
            await asyncio.sleep(.5)
        await scheduler.close()

        results = [
            await queue.get()
            for _ in range(queue.qsize())
        ]
    results.sort(key=lambda res: res[0])
    return [
        html for p, html in results
    ]


def get_htmls(url: str,
              start: int = 0,
              stop: int = 1,
              **kwargs) -> List[str]:
    """ Run coro, get html codes of the pages.

    URLs will be created for i in range(start, stop),
     HTTP tag 'p' (page) is i.

    :param url: str, URL.
    :param start: int, start page index.
    :param stop: int, stop page index.
    :param kwargs: HTTP tags.
    :return: list of str, html codes of the pages.
    """
    logger.info(f"Requested: ({start};{stop}), with params {kwargs}")
    html_codes = asyncio.run(
        get_htmls_coro(url, start, stop, **kwargs)
    )
    logger.info("Request was successfully completed ")
    return html_codes


def is_http_request_correct(url: str,
                            **kwargs) -> bool:
    """ Whether the request is correct.
    It's correct If there's no exception catch.

    :param url: str, request url.
    :param kwargs: request HTTP tags.
    :return: bool, whether the request is correct.
    """
    # coro writes logs by itself
    try:
        htmls = get_htmls(url, **kwargs)
    except Exception:
        return False
    return bool(htmls)


def whether_result_found(url: str,
                         **kwargs) -> bool:
    """ Whether the page contains results.

    :param url: str, request url.
    :param kwargs: request HTTP tags.
    :exception RuntimeError: if HTTP request was wrong.
    """
    logger.info("Validating that the request is OK")
    try:
        page_html = get_htmls(url, **kwargs)[0]
    except Exception:
        logger.exception("The request is not correct")
        raise RuntimeError
    logger.info("The request is correct")

    logger.info("Validating that the result exits")
    soup = bs4.BeautifulSoup(page_html, 'lxml')

    # TODO: сузить круг поиска
    content = soup.find('div', {'class': 'content'}).text
    res_msg = ('По этому запросу ничего не найдено.' in content or
               'No results match the search query.' in content)
    return not res_msg


def does_page_exist(url: str,
                    p_index: int,
                    **kwargs) -> bool:
    """ Whether a page at the index exists.

    It means, the number of the page in 'pager' is equal to expected index.
    RNC redirects to the first page if the page at the number doesn't exist.
    Here it's assumed, that the request's correct.

    :param url: str, URL.
    :param p_index: int, index of page. Indexing starts with 0.
    :param kwargs: HTTP tags.
    :return: bool, whether a page exists.
    """
    # indexing starts with 0
    start = p_index
    start = start * (start >= 0)
    stop = p_index + 1

    # request's correct → first page exists
    if stop is 1:
        return True

    page_html = get_htmls(url, start, stop, **kwargs)[0]
    soup = bs4.BeautifulSoup(page_html, 'lxml')

    pager = soup.find('p', {'class': 'pager'})
    if pager:
        p_num = pager.b
        if not p_num:
            return False
        # page index from pager should be equal to expected index
        return p_num.text == str(stop)

    # if there's no pager, but result exists.
    # this might happen if expand=full or out=kwic
    first_page_code = get_htmls(url, 0, 1, **kwargs)[0]
    return page_html != first_page_code


def is_request_correct(url: str,
                       p_count: int,
                       **kwargs) -> bool:
    """ Check:
        – is the HTTP request correct (means there're no exceptions catch).

        – has there been any result.

        – does a page at the number exist (
        means RNC doesn't redirect to the first page).

    :param url: str, request url.
    :param p_count: int, request count of pages.
    :param kwargs: request HTTP tags.
    :return: True if everything's OK, an exception otherwise.
    :exception ValueError: something's wrong.
    """
    logger.info("Validating that everything is OK")
    try:
        # to reduce the number of requests
        # the two checks are combined into one.
        # coro writes logs by itself
        assert whether_result_found(url, **kwargs) is True
    except AssertionError:
        logger.exception("HTTP request is OK, but no result found")
        raise ValueError("No result found")
    except RuntimeError:
        logger.exception("HTTP request is wrong")
        raise ValueError("Wrong HTTP request")
    else:
        logger.info("HTTP request is correct, result found")

    logger.info("Validating that the last page exists")
    if not does_page_exist(url, p_count - 1, **kwargs):
        logger.error("Everything is OK, but last page doesn't exist")
        raise ValueError("Last page doesn't exist")
    logger.info("The last page exists")

    logger.info("Validated successfully")
    return True


async def fetch_download(url: str,
                         ses: aiohttp.ClientSession,
                         filename: str,
                         **kwargs) -> None:
    """ Coro, downloading and writing media file from RNC.

    If response status == 429 sleep 24s and try again.

    :param url: str, file's url.
    :param ses: aiohttp.ClientSession.
    :param filename: str, name of the file.
    :param kwargs: HTTP tags to request.
    :return: None.
    :exception: all exceptions should be processed here.
    """
    try:
        resp = await ses.get(url, allow_redirects=True, params=kwargs)
    except Exception:
        logger.exception("Cannot get answer from RNC")
        return

    try:
        if resp.status != 429:
            resp.raise_for_status()
    except Exception:
        logger.error(
            f"{resp.status}: {resp.reason} requesting to {resp.url}"
        )
        resp.close()
        return

    if resp.status == 200:
        content = await resp.read()
        async with aiofiles.open(filename, 'wb') as f:
            await f.write(content)
        resp.close()
    elif resp.status == 429:
        logger.debug(
            f"429, {resp.reason} downloading '{filename}', wait {WAIT}s"
        )
        resp.close()
        await asyncio.sleep(WAIT)
        await fetch_download(url, ses, filename)
    else:
        logger.critical("This case must not happened")


async def download_docs_coro(url_to_name: List[Tuple[str, str]]) -> None:
    """ Coro executing fetch_download coro, catching exceptions.

    :param url_to_name: list of tuples of str, pairs: url – filename.
    :return None.
    """
    timeout = aiohttp.ClientTimeout(WAIT)
    async with aiohttp.ClientSession(timeout=timeout) as ses:
        scheduler = await aiojobs.create_scheduler(limit=None)
        for url, filename in url_to_name:
            await scheduler.spawn(
                fetch_download(url, ses, filename)
            )

        while len(scheduler) != 0:
            await asyncio.sleep(.5)

        await scheduler.close()


def download_docs(url_to_name: List[Tuple[str, str]]) -> None:
    """ Run coro, download the files.

    :param url_to_name: list of tuples of str, pairs: url – filename.
    :return: None.
    """
    logger.info(f"Requested {len(url_to_name)} files to download")
    asyncio.run(download_docs_coro(url_to_name))
    logger.info(f"Downloading successfully completed")



import os
from pathlib import Path
from time import sleep

import pytest

import rnc.corpora as rnc


class TemplateCorpusTest:
    corp_type = rnc.Corpus

    corp_normal_obj = None
    corp_kwic_obj = None

    full_query = {
        'ты': {
            'gramm': {
                'case': ['acc', 'nom', 'gen'],
                'num': ['sg', 'pl']
            },
            'flags': {
                'position': ['amark', 'bmark'],
            }
        }
    }

    #########################
    #    Test requesting    #
    #########################

    def test_empty_query(self):
        with pytest.raises(ValueError):
            self.corp_type('', 1, marker=str.upper)

    def test_p_count_zero(self):
        with pytest.raises(ValueError):
            self.corp_type('корпус', 0, marker=None)

    def test_one_str_with_one_word(self):
        corp = self.corp_type('ты', 1, marker=str.capitalize, spd=1, dpp=5)
        corp.request_examples()

        assert len(corp) > 1
        sleep(5)

    def test_one_form_without_gram(self):
        corp = self.corp_type({'ты': ''}, 1, marker=str.capitalize, spd=1, dpp=5)
        corp.request_examples()

        assert len(corp) > 1
        sleep(5)

    def test_one_form_with_one_gram(self):
        corp = self.corp_type({'ты': 'nom'}, 1, marker=str.capitalize, spd=1)
        corp.request_examples()

        assert len(corp) > 1
        sleep(5)

    def test_one_form_with_grams_and_flags(self):
        query = {
            'ты': {
                'gramm': 'acc',
                'flags': 'amark'
            }
        }
        corp = self.corp_type(query, 1, marker=str.capitalize)
        corp.request_examples()

        assert len(corp) > 1
        sleep(5)

    def test_one_form_with_several_grams_and_flags(self):
        corp = self.corp_type(self.full_query, 1, marker=str.capitalize)
        corp.request_examples()

        assert len(corp) > 1
        sleep(5)

    def test_full_query_with_lexform(self):
        corp = self.corp_type('слово бога', 1, text='lexform')
        corp.request_examples()

        assert len(corp) > 1
        sleep(5)

    def test_full_query_with_kwic_without_kwsz(self):
        corp = self.corp_type(self.full_query, 1, marker=str.capitalize, out='kwic')
        corp.request_examples()

        assert len(corp) > 1
        sleep(5)

    def test_full_query_with_kwic_with_kwsz(self):
        corp = self.corp_type(self.full_query, 1, marker=str.capitalize, out='kwic', kwsz=7)
        corp.request_examples()

        assert len(corp) > 1
        sleep(5)

    def test_sort_in_request(self):
        default = self.corp_type('ты', 1)
        default.request_examples()
        sleep(5)

        by_creation_date = self.corp_type('ты', 1, sort='i_grcreated_inv')
        by_creation_date.request_examples()
        sleep(5)

        assert default.data != by_creation_date.data

    def test_subcorpus(self):
        corp = self.corp_type(
            'ты', 1,
            subcorpus="JSONeyJkb2Nfc2V4IjogWyLQvNGD0LYiXSwgImRvY19pX3RhZ2dpbmciOiBbIjEiXX0%3D"
        )
        corp.request_examples()

        assert len(corp) >= 1
        sleep(5)

    def test_call(self):
        corp = self.corp_type('ты', 1, marker=str.capitalize, out='kwic')
        corp()

        assert len(corp) > 1

    #########################
    #   Working with file   #
    #########################

    def test_dump_normal(self):
        self.corp_normal_obj.dump()

    def test_dump_kwic(self):
        self.corp_kwic_obj.dump()

    def test_load_normal(self):
        corp = self.corp_type(file=self.corp_normal_obj.file)

        assert len(corp) == len(self.corp_normal_obj)
        assert (
            from_file == from_corp
            for from_file, from_corp in zip(corp, self.corp_normal_obj)
        )

    def test_load_kwic(self):
        corp = self.corp_type(file=self.corp_kwic_obj.file)

        assert len(corp) == len(self.corp_kwic_obj)
        assert (
            from_file == from_corp
            for from_file, from_corp in zip(corp, self.corp_kwic_obj)
        )

    def test_load_to_wrong_corpus(self):
        with pytest.raises(NotImplementedError):
            self.corp_type(file='data\\wrong_mode.csv')

    def test_request_if_base_loaded(self):
        corp = self.corp_type(file=self.corp_normal_obj.file)
        with pytest.raises(RuntimeError):
            corp.request_examples()

    def test_wrong_filetype(self):
        with pytest.raises(TypeError):
            corp = self.corp_type('ты', 1, file='file.txt')

    def test_default_filetype(self):
        corp = self.corp_type('ты', 1)

        assert isinstance(corp.file, Path) and corp.file.suffix == '.csv'

    def test_equality_wordforms_from_rnc_and_from_file(self):
        corp = self.corp_type(file=self.corp_kwic_obj.file)

        assert all(
            from_file == from_corp
            for from_file, from_corp in zip(corp, self.corp_kwic_obj)
        )

    #########################
    #    Test properties    #
    #########################

    def test_data_type(self):
        assert isinstance(self.corp_normal_obj.data, list)

    def test_data_elements_type(self):
        assert all(
            isinstance(ex, self.corp_normal_obj.ex_type)
            for ex in self.corp_normal_obj
        )

    def test_query_dict(self):
        corp = self.corp_type({'ты': 'acc'}, 1)

        assert isinstance(corp.query, dict) and corp.query

    def test_query_str(self):
        corp = self.corp_type('ты', 1)

        assert isinstance(corp.query, str) and corp.query

    def test_forms_in_query_dict(self):
        corp = self.corp_type({'ты': 'acc', 'готов': {}}, 1)

        assert isinstance(corp.forms_in_query, list)
        assert all(isinstance(form, str) for form in corp.forms_in_query)
        assert len(corp.forms_in_query) is 2
        assert corp.forms_in_query == ['ты', 'готов']

    def test_forms_in_query_str(self):
        corp = self.corp_type('ты готов ', 1)

        assert isinstance(corp.forms_in_query, list)
        assert all(isinstance(form, str) for form in corp.forms_in_query)
        assert len(corp.forms_in_query) is 2
        assert corp.forms_in_query == ['ты', 'готов']

    def test_p_count(self):
        corp = self.corp_type('ты', 1)

        assert corp.p_count is 1

    def test_found_wordforms_from_file(self):
        corp = self.corp_type(file=self.corp_normal_obj.file)

        assert corp.found_wordforms == self.corp_normal_obj.found_wordforms

    def test_url(self):
        corp = self.corp_type('ты', 1)

        assert isinstance(corp.url, str) and corp.url

    def test_amount_of_docs_normal(self):
        assert isinstance(self.corp_normal_obj.amount_of_docs, int)

    def test_amount_of_docs_kwic(self):
        assert self.corp_kwic_obj.amount_of_docs is None

    def test_amount_of_contexts_normal(self):
        assert isinstance(self.corp_normal_obj.amount_of_contexts, int)

    def test_amount_of_contexts_kwic(self):
        assert isinstance(self.corp_kwic_obj.amount_of_contexts, int)

    ##########################
    # Test working with data #
    ##########################

    def test_open_url(self):
        self.corp_normal_obj.open_url()

    def test_open_graphic(self):
        self.corp_normal_obj.open_graphic()

    def test_copy(self):
        copy = self.corp_normal_obj.copy()

        assert copy.data == self.corp_normal_obj.data

    def test_sort_data(self):
        copy = self.corp_normal_obj.copy()
        copy.sort_data(key=lambda x: len(x.txt))

        assert copy.data != self.corp_normal_obj.data

    def test_pop(self):
        copy = self.corp_normal_obj.copy()
        example = copy.pop(0)

        assert any(
            example == ex
            for ex in self.corp_normal_obj)
        assert all(
            example != ex
            for ex in copy
        )

    def test_contains(self):
        copy = self.corp_normal_obj.copy()
        example = copy.pop(0)

        assert (example in self.corp_normal_obj and
                example not in copy)

    def test_shuffle(self):
        copy = self.corp_normal_obj.copy()
        copy.shuffle()

        assert copy.data != self.corp_normal_obj.data

    def test_clear(self):
        copy = self.corp_normal_obj.copy()
        copy.clear()

        assert not copy.data
        assert copy.query
        assert copy.p_count
        assert copy.params

    def test_filter(self):
        copy = self.corp_normal_obj.copy()
        copy.filter(lambda x: x.txt is None)

        assert len(copy) == 0

    def test_getattr_normal(self):
        mode = self.corp_normal_obj.mode

        assert isinstance(mode, str) and mode

    def test_getattr_none(self):
        assert self.corp_normal_obj.name is None

    def test_getitem_one(self):
        item = self.corp_normal_obj[0]

        assert item in self.corp_normal_obj

    def test_getitem_slice(self):
        sliced = self.corp_normal_obj[::-1]

        assert all(
            lhs == rhs
            for lhs, rhs in zip(self.corp_normal_obj.data[::-1], sliced)
        )

    def test_delitem(self):
        copy = self.corp_normal_obj.copy()
        del copy[:]

        assert copy == 0

    ##########################
    #        Test <=>        #
    ##########################

    def test_lt_with_int(self):
        assert self.corp_normal_obj < 10
        assert self.corp_kwic_obj < 10

    def test_lt_with_corp(self):
        copy = self.corp_normal_obj.copy()
        del copy[0]

        assert copy < self.corp_normal_obj

    def test_le_with_int(self):
        assert self.corp_normal_obj <= 5
        assert self.corp_kwic_obj <= 5

    def test_le_with_corp(self):
        copy = self.corp_normal_obj.copy()
        del copy[0]

        assert copy <= self.corp_normal_obj

    def test_eq_with_int(self):
        assert self.corp_normal_obj == len(self.corp_normal_obj)
        assert self.corp_kwic_obj == len(self.corp_kwic_obj)

    def test_eq_with_corp(self):
        copy = self.corp_normal_obj.copy()

        assert copy == self.corp_normal_obj

    def test_ne_with_int(self):
        assert self.corp_normal_obj != 0
        assert self.corp_kwic_obj != 0

    def test_ne_with_corp(self):
        copy = self.corp_normal_obj.copy()
        del copy[0]

        assert copy != self.corp_normal_obj

    def test_gt_with_int(self):
        assert self.corp_normal_obj > 0
        assert self.corp_kwic_obj > 0

    def test_gt_with_corp(self):
        copy = self.corp_normal_obj.copy()
        del copy[0]

        assert self.corp_normal_obj > copy

    def test_ge_with_int(self):
        assert self.corp_normal_obj >= 1
        assert self.corp_kwic_obj >= 1

    def test_ge_with_corp(self):
        copy = self.corp_normal_obj.copy()
        del copy[0]

        assert self.corp_normal_obj >= copy

    ##########################
    #   Test class setters   #
    ##########################

    def test_set_spd_normal(self):
        self.corp_type.set_spd(20)
        corp = self.corp_type('ты', 1)

        assert corp.spd is 20

    def test_set_spd_exception(self):
        with pytest.raises(TypeError):
            self.corp_type.set_spd('12')

    def test_set_dpp_normal(self):
        self.corp_type.set_dpp(20)
        corp = self.corp_type('ты', 1)

        assert corp.dpp is 20

    def test_set_dpp_exception(self):
        with pytest.raises(TypeError):
            self.corp_type.set_dpp('12')

    def test_set_text_normal(self):
        self.corp_type.set_text('lexform')
        corp = self.corp_type('ты', 1)

        assert corp.text == 'lexform'

    def test_set_text_exception(self):
        with pytest.raises(TypeError):
            self.corp_type.set_text(12)

    def test_set_sort_normal(self):
        self.corp_type.set_sort('i_grcreated_inv')
        corp = self.corp_type('ты', 1)

        assert corp.sort == 'i_grcreated_inv'

    def test_set_sort_exception(self):
        with pytest.raises(TypeError):
            self.corp_type.set_sort(12)

    def test_set_min_normal(self):
        self.corp_type.set_min(10)
        corp = self.corp_type('ты готов', 1)

        assert corp.min2 is 10

    def test_set_min_exception(self):
        with pytest.raises(TypeError):
            self.corp_type.set_min('12')

    def test_set_max_normal(self):
        self.corp_type.set_max(10)
        corp = self.corp_type('ты готов', 1)

        assert corp.max2 is 10

    def test_set_max_exception(self):
        with pytest.raises(TypeError):
            self.corp_type.set_max('12')

    def test_set_restrict_show_exception_str(self):
        with pytest.raises(TypeError):
            self.corp_type.set_restrict_show('False')

    def test_set_restrict_show_exception_list(self):
        with pytest.raises(TypeError):
            self.corp_type.set_restrict_show([False])


class TestMainCorpus(TemplateCorpusTest):
    corp_type = rnc.MainCorpus

    # for test working with file
    corp_normal_obj = corp_type('ты', 1, dpp=5, spd=1)
    corp_kwic_obj = corp_type('ты', 1, dpp=5, spd=1, out='kwic')

    corp_normal_obj.request_examples()
    sleep(5)
    corp_kwic_obj.request_examples()
    sleep(5)


class TestPaper2000Corpus(TemplateCorpusTest):
    corp_type = rnc.Paper2000Corpus

    corp_normal_obj = corp_type('ты', 1, dpp=5, spd=1)
    corp_kwic_obj = corp_type('ты', 1, dpp=5, spd=1, out='kwic')

    corp_normal_obj.request_examples()
    sleep(5)
    corp_kwic_obj.request_examples()
    sleep(5)

    def test_subcorpus(self):
        corp = self.corp_type(
            'ты', 1,
            subcorpus="JSONeyJkb2NfaV9sZV9zdGFydF95ZWFyIjogWyIyMDEwIl19"
        )
        corp.request_examples()

        assert len(corp) >= 1
        sleep(5)


class TestPaperRegionalCorpus(TemplateCorpusTest):
    corp_type = rnc.PaperRegionalCorpus

    corp_normal_obj = corp_type('ты', 1, dpp=5, spd=1)
    corp_kwic_obj = corp_type('ты', 1, dpp=5, spd=1, out='kwic')

    corp_normal_obj.request_examples()
    sleep(5)
    corp_kwic_obj.request_examples()
    sleep(5)

    def test_full_query_with_lexform(self):
        corp = self.corp_type('Владимир Путин', 1, text='lexform')
        corp.request_examples()

        assert len(corp) > 1
        sleep(5)

    def test_subcorpus(self):
        corp = self.corp_type(
            'ты', 1,
            subcorpus="JSONeyJkb2NfaV9sZV9zdGFydF95ZWFyIjogWyIyMDEwIl19"
        )
        corp.request_examples()

        assert len(corp) >= 1
        sleep(5)

    def test_open_graphic(self):
        with pytest.raises(RuntimeError):
            self.corp_normal_obj.open_graphic()


class TestParallelCorpus(TemplateCorpusTest):
    corp_type = rnc.ParallelCorpus

    corp_normal_obj = corp_type('ты', 1, dpp=5, spd=1)
    corp_kwic_obj = corp_type('ты', 1, dpp=5, spd=1, out='kwic')

    corp_normal_obj.request_examples()
    sleep(5)
    corp_kwic_obj.request_examples()
    sleep(5)

    def test_full_query_with_lexform(self):
        corp = self.corp_type('Владимир Путин', 1, text='lexform')
        corp.request_examples()

        assert len(corp) > 1
        sleep(5)

    def test_subcorpus(self):
        corp = self.corp_type(
            'ты', 1,
            subcorpus='JSONeyJkb2NfaV9sZV9zdGFydF95ZWFyIjogWyIyMDAwIl19'
        )
        corp.request_examples()

        assert len(corp) >= 1
        sleep(5)

    def test_sort_data(self):
        copy = self.corp_normal_obj.copy()
        copy.sort_data(key=lambda x: len(x.ru))

        assert copy.data != self.corp_normal_obj.data

    def test_open_graphic(self):
        with pytest.raises(RuntimeError):
            self.corp_normal_obj.open_graphic()


class TestMultilingualParaCorpus(TemplateCorpusTest):
    corp_type = rnc.MultilingualParaCorpus

    corp_normal_obj = corp_type('ты', 1, dpp=5, spd=1)
    corp_kwic_obj = corp_type('ты', 1, dpp=5, spd=1, out='kwic')

    corp_normal_obj.request_examples()
    sleep(5)
    corp_kwic_obj.request_examples()
    sleep(5)

    def test_full_query_with_lexform(self):
        corp = self.corp_type('ты готов', 1, text='lexform')
        corp.request_examples()

        assert len(corp) > 1
        sleep(5)

    def test_subcorpus(self):
        pass
        # it's impossible to set subcorpus in MultilingualParaCorpus

        # corp = self.corp_type(
        #     'ты', 1,
        #     subcorpus='JSONeyJkb2NfaV9sZV9zdGFydF95ZWFyIjogWyIyMDAwIl19'
        # )
        # corp.request_examples()
        #
        # assert len(corp) >= 1
        # sleep(5)

    def test_sort_data(self):
        copy = self.corp_normal_obj.copy()
        copy.sort_data(key=lambda x: len(x.src))

        assert copy.data != self.corp_normal_obj.data

    def test_dump_normal(self):
        with pytest.raises(NotImplementedError):
            super().test_dump_normal()

    def test_dump_kwic(self):
        with pytest.raises(NotImplementedError):
            super().test_dump_kwic()

    def test_load_kwic(self):
        pass

    def test_load_normal(self):
        pass

    def test_load_to_wrong_corpus(self):
        pass

    def test_request_if_base_loaded(self):
        pass

    def test_equality_wordforms_from_rnc_and_from_file(self):
        pass

    def test_found_wordforms_from_file(self):
        pass

    def test_open_graphic(self):
        with pytest.raises(RuntimeError):
            self.corp_normal_obj.open_graphic()


class TestTutoringCorpus(TemplateCorpusTest):
    corp_type = rnc.TutoringCorpus

    corp_normal_obj = corp_type('ты', 1, dpp=5, spd=1)
    corp_kwic_obj = corp_type('ты', 1, dpp=5, spd=1, out='kwic')

    corp_normal_obj.request_examples()
    sleep(5)
    corp_kwic_obj.request_examples()
    sleep(5)

    def test_full_query_with_lexform(self):
        corp = self.corp_type('Владимир Путин', 1, text='lexform')
        corp.request_examples()

        assert len(corp) >= 1
        sleep(5)

    def test_open_graphic(self):
        with pytest.raises(RuntimeError):
            self.corp_normal_obj.open_graphic()


class TestDialectalCorpus(TemplateCorpusTest):
    corp_type = rnc.DialectalCorpus

    corp_normal_obj = corp_type('ты', 1, dpp=5, spd=1)
    corp_kwic_obj = corp_type('ты', 1, dpp=5, spd=1, out='kwic')

    corp_normal_obj.request_examples()
    sleep(5)
    corp_kwic_obj.request_examples()
    sleep(5)

    def test_full_query_with_lexform(self):
        corp = self.corp_type('дак ты', 1, text='lexform')
        corp.request_examples()

        assert len(corp) > 1
        sleep(5)

    def test_subcorpus(self):
        corp = self.corp_type(
            'ты', 1,
            subcorpus='JSONeyJkb2NfcmVnaW9uIjogWyLQmtCw0YDQtdC70LjRjyJdfQ%3D%3D'
        )
        corp.request_examples()

        assert len(corp) >= 1
        sleep(5)

    def test_open_graphic(self):
        with pytest.raises(RuntimeError):
            self.corp_normal_obj.open_graphic()


class TestSpokenCorpus(TemplateCorpusTest):
    corp_type = rnc.SpokenCorpus

    corp_normal_obj = corp_type('ты', 1, dpp=5, spd=1)
    corp_kwic_obj = corp_type('ты', 1, dpp=5, spd=1, out='kwic')

    corp_normal_obj.request_examples()
    sleep(5)
    corp_kwic_obj.request_examples()
    sleep(5)

    def test_open_graphic(self):
        with pytest.raises(RuntimeError):
            self.corp_normal_obj.open_graphic()


class TestAccentologicalCorpus(TemplateCorpusTest):
    corp_type = rnc.AccentologicalCorpus

    corp_normal_obj = corp_type('ты', 1, dpp=5, spd=1)
    corp_kwic_obj = corp_type('ты', 1, dpp=5, spd=1, out='kwic')

    corp_normal_obj.request_examples()
    sleep(5)
    corp_kwic_obj.request_examples()
    sleep(5)

    def test_subcorpus(self):
        corp = self.corp_type(
            'ты', 1,
            subcorpus='JSONeyJkb2NfYXV0aG9yIjogWyLQkC7QoS4g0J_Rg9GI0LrQuNC9Il0sICJkb2NfaV9sZV9lbmRfeWVhciI6IFsiMTgzMCJdfQ%3D%3D'
        )
        corp.request_examples()

        assert len(corp) >= 1
        sleep(5)

    def test_open_graphic(self):
        with pytest.raises(RuntimeError):
            self.corp_normal_obj.open_graphic()


class TestMultimodalCorpus(TemplateCorpusTest):
    corp_type = rnc.MultimodalCorpus

    corp_normal_obj = corp_type('корпус', 1, dpp=5, spd=1)
    corp_kwic_obj = corp_type('корпус', 1, dpp=5, spd=1, out='kwic')

    corp_normal_obj.request_examples()
    sleep(5)
    corp_kwic_obj.request_examples()
    sleep(5)

    def test_download_all(self):
        self.corp_normal_obj.download_all()
        files = os.listdir(self.corp_normal_obj.MEDIA_FOLDER)

        assert all(
            ex.filepath.name in files
            for ex in self.corp_normal_obj
        )


import pathlib
import re
import sys

from setuptools import setup, find_packages


if sys.version_info < (3, 7):
    raise RuntimeError('Python >= 3.7 is required')

HERE = pathlib.Path(__file__).parent
txt = (HERE / 'rnc' / '__init__.py').read_text('utf-8')

try:
    version = re.findall(
        r"^__version__ = '([^']+)'\r?$", txt, re.M)[0]
except IndexError:
    raise RuntimeError('Unable to determine version.')


setup(
    name='rnc',
    version=version,
    url='https://github.com/kunansy/RNC',
    packages=find_packages(exclude='tests'),
    python_requires='>=3.7',
    license='MIT',
    install_requires=['bs4',
                      'beautifulsoup4>=4.9',
                      'aiohttp>=3.6',
                      'lxml>=4.5',
                      'aiofiles>=0.5',
                      'aiojobs>=0.3',
                      'lxml'],
    author="Kolobov Kirill",
    author_email='alniconim@gmail.com',
    description='API for Russian National Corpus',
    long_description=open('README.md', encoding='utf-8').read(),
    long_description_content_type='text/markdown',
    zip_safe=False,
)



from typing import Any


class Page:
    @classmethod
    def e(cls,
          value: int) -> Any:
        """" Get amount of pages exactly equals to the value """
        pass

    @classmethod
    def le(cls,
           value: int) -> Any:
        """ Get amount of pages <= than value """
        pass


class Subcorpus:
    def __getitem__(self, item: str) -> str:
        """ Get attribute from Person or from Parallel.
        Raise KeyError if there is no the item in both them.

        :param item: str, key name.
        :return: str, key value.
        :exception KeyError: if the key doesn't exist.
        """
        try:
            return getattr(self.Person, item)
        except AttributeError:
            pass
        try:
            return getattr(self.Parallel, item)
        except AttributeError:
            raise KeyError(f"No key '{item}' in either Person or Parallel")

    def __getattr__(self, item: str) -> str:
        """ Get attribute from Person or from Parallel.
        Raise AttributeError if there is no the item in both them.

        :param item: str, key name.
        :return: str, key value.
        :exception AttributeError: if the key doesn't exist.
        """
        try:
            return getattr(super(), item)
        except AttributeError:
            pass
        try:
            return self[item]
        except KeyError as e:
            raise AttributeError(e)

    class Person:
        Pushkin = 'JSONeyJkb2NfYXV0aG9yIjogWyLQkC7QoS4g0J_Rg9GI0LrQuNC9Il19'
        Dostoyevsky = 'JSONeyJkb2NfYXV0aG9yIjogWyLQpC7QnC4g0JTQvtGB0YLQvtC10LLRgdC60LjQuSJdfQ%3D%3D'
        TolstoyLN = 'JSONeyJkb2NfYXV0aG9yIjogWyLQmy7QnS4g0KLQvtC70YHRgtC-0LkiXX0%3D'
        Chekhov = 'JSONeyJkb2NfYXV0aG9yIjogWyLQkC7Qny4g0KfQtdGF0L7QsiJdfQ%3D%3D'
        Gogol = 'JSONeyJkb2NfYXV0aG9yIjogWyLQnS7Qki4g0JPQvtCz0L7Qu9GMIl19'
        Turgenev = 'JSONeyJkb2NfYXV0aG9yIjogWyLQmC7QoS4g0KLRg9GA0LPQtdC90LXQsiJdfQ%3D%3D'

        def __getitem__(self, item: str) -> str:
            """ Get attribute from Person.
            Raise KeyError if there is no the item.

            :param item: str, key name.
            :return: str, key value.
            :exception KeyError: if the key doesn't exist.
            """
            try:
                return getattr(self, item)
            except AttributeError as e:
                raise KeyError(e)

    class Parallel:
        English = en = 'JSONeyJkb2NfbGFuZyI6IFsiZW5nIl0sICJpc19wYXJhX2JvdGhfcGFpcnMiOiBbdHJ1ZV19'
        Armenian = arm = 'JSONeyJkb2NfbGFuZyI6IFsiYXJtIl0sICJpc19wYXJhX2JvdGhfcGFpcnMiOiBbdHJ1ZV19'
        Bashkir = bas = 'JSONeyJkb2NfbGFuZyI6IFsiYmFzaCJdLCAiaXNfcGFyYV9ib3RoX3BhaXJzIjogW3RydWVdfQ=='
        Belarusian = bel = 'JSONeyJkb2NfbGFuZyI6IFsiYmVsIl0sICJpc19wYXJhX2JvdGhfcGFpcnMiOiBbdHJ1ZV19'
        Bulgarian = bul = 'JSONeyJkb2NfbGFuZyI6IFsiYnVsIl0sICJpc19wYXJhX2JvdGhfcGFpcnMiOiBbdHJ1ZV19'
        Buryatian = bur = 'JSONeyJkb2NfbGFuZyI6IFsiYnVhIl0sICJpc19wYXJhX2JvdGhfcGFpcnMiOiBbdHJ1ZV19'
        Spanish = sp = 'JSONeyJkb2NfbGFuZyI6IFsiZXNwIl0sICJpc19wYXJhX2JvdGhfcGFpcnMiOiBbdHJ1ZV19'
        Italian = it = 'JSONeyJkb2NfbGFuZyI6IFsiaXRhIl0sICJpc19wYXJhX2JvdGhfcGFpcnMiOiBbdHJ1ZV19'
        Chinese = ch = 'JSONeyJkb2NfbGFuZyI6IFsiemhvIl0sICJpc19wYXJhX2JvdGhfcGFpcnMiOiBbdHJ1ZV19'
        Latvian = lat = 'JSONeyJkb2NfbGFuZyI6IFsibGF2Il0sICJpc19wYXJhX2JvdGhfcGFpcnMiOiBbdHJ1ZV19'
        Lithuanian = lit = 'JSONeyJkb2NfbGFuZyI6IFsibGl0Il0sICJpc19wYXJhX2JvdGhfcGFpcnMiOiBbdHJ1ZV19'
        German = ger = 'JSONeyJkb2NfbGFuZyI6IFsiZ2VyIl0sICJpc19wYXJhX2JvdGhfcGFpcnMiOiBbdHJ1ZV19'
        Polish = pol = 'JSONeyJkb2NfbGFuZyI6IFsicG9sIl0sICJpc19wYXJhX2JvdGhfcGFpcnMiOiBbdHJ1ZV19'
        Ukrainian = ukr = 'JSONeyJkb2NfbGFuZyI6IFsidWtyIl0sICJpc19wYXJhX2JvdGhfcGFpcnMiOiBbdHJ1ZV19'
        French = fr = 'JSONeyJkb2NfbGFuZyI6IFsiZnJhIl0sICJpc19wYXJhX2JvdGhfcGFpcnMiOiBbdHJ1ZV19'
        Finnish = fin = 'JSONeyJkb2NfbGFuZyI6IFsiZmluIl0sICJpc19wYXJhX2JvdGhfcGFpcnMiOiBbdHJ1ZV19'
        Czech = cz = 'JSONeyJkb2NfbGFuZyI6IFsiY3plIl0sICJpc19wYXJhX2JvdGhfcGFpcnMiOiBbdHJ1ZV19'
        Swedish = sw = 'JSONeyJkb2NfbGFuZyI6IFsic3ZlIl0sICJpc19wYXJhX2JvdGhfcGFpcnMiOiBbdHJ1ZV19'
        Estonian = es = 'JSONeyJkb2NfbGFuZyI6IFsiZXN0Il0sICJpc19wYXJhX2JvdGhfcGFpcnMiOiBbdHJ1ZV19'

        def __getitem__(self, item: str) -> str:
            """ Get attribute from Parallel.
            Raise KeyError if there is no the item.

            :param item: str, key name.
            :return: str, key value.
            :exception KeyError: if the key doesn't exist.
            """
            try:
                return getattr(self, item)
            except AttributeError as e:
                raise KeyError(e)
